# coding: utf-8

from numpy import trapz
from typing import Union, Dict
import logging

from .meta_path import Path, ConditionType


class Integrator(Path):
  """This Path integrates an incoming label over time and returns the
  integration as an output signal.

  Let `f(t)` be the input signal, `v(t)` the value of the output, `m` the
  inertia and `t0` the timestamp of the beginning of this Path.

  Then the output value for this Path will be
  :math:`v(t) = v(t0) - [I(t0 -> t)f(t)dt] / m`.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Inertia to Integrator
  """

  def __init__(self,
               condition: Union[str, ConditionType],
               inertia: float,
               func_label: str,
               time_label: str = 't(s)',
               init_value: float = None) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      condition: The condition for switching to the next Path. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      inertia: Value of the equivalent inertia to use for driving the signal.
        In the above formula, it is the value of `m`. The larger this value,
        the slower the changes in the signal value.
      func_label: The name of the label of the input value to integrate.
      time_label: The name of the time label for the integration.
      init_value: If given, overwrites the last value of the signal as the
        starting point for the inertia path. In the specific case when this
        path is the first one in the Generator Paths, this argument must be
        given !
    
    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionchanged:: 1.5.10 renamed *flabel* argument to *func_label*
    .. versionchanged:: 1.5.10 renamed *tlabel* argument to *time_label*
    .. versionchanged:: 1.5.10 renamed *value* argument to *init_value*
    .. versionremoved:: 1.5.10 *const* argument
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    super().__init__()

    if init_value is None and self.last_cmd is None:
      raise ValueError('For the first path, an init_value must be given !')

    # Setting the attributes
    self._condition = self.parse_condition(condition)
    self._time_label = time_label
    self._func_label = func_label
    self._inertia = inertia
    self._value = self.last_cmd if init_value is None else init_value
    self._last_t = None
    self._last_val = None

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Gets the latest values of the incoming label, integrates them and
    changes the output accordingly.

    Also raises :exc:`StopIteration` in case the stop condition is met.
    """

    # Checking if the stop condition is met
    if self._condition(data):
      self.log(logging.DEBUG, "Stop condition met")
      raise StopIteration

    if self._time_label in data and self._func_label in data:
      # Getting the last values from the received data
      times = data[self._time_label]
      values = data[self._func_label]

      # Including the last values from the last loop
      if self._last_val is not None and self._last_t is not None:
        times = [self._last_t] + times
        values = [self._last_val] + values

      # Keeping in memory the last values from this loop
      if times and values:
        self._last_t = times[-1]
        self._last_val = values[-1]

      # Performing the integration and subtracting from the previous value
      self._value -= trapz(values, times) / self._inertia

    # Returning the current value
    return self._value



# coding: utf-8

from multiprocessing import Process, RLock, Event, Value, get_start_method
import multiprocessing.synchronize
import multiprocessing.context
import multiprocessing.queues
from multiprocessing.sharedctypes import Synchronized
import signal
from _io import FileIO
from tempfile import TemporaryFile
from typing import List, Dict, Any, Optional, Tuple
from contextlib import contextmanager
from dataclasses import dataclass
import logging
import logging.handlers

from ..._global import OptionalModule
try:
  from usb.core import find, Device, USBTimeoutError
  from usb import util
except (FileNotFoundError, ModuleNotFoundError):
  find = OptionalModule('pyusb')
  Device = OptionalModule
  USBTimeoutError = OptionalModule('pyusb')
  util = OptionalModule('pyusb')

USBArgsType = Tuple[int, multiprocessing.synchronize.RLock, FileIO,
                    FileIO, multiprocessing.synchronize.RLock,
                    Synchronized]


@dataclass
class BlockObjects:
  """This class stores all the objects specific to a single Block, in order to
  easily pass them to the :class:`USBServer` process.

  Such a class will be created for each :class:`~crappy.blocks.Block`
  registering with the :class:`USBServer`.
  """

  ser_num: str
  lock: multiprocessing.synchronize.RLock
  device: Device

  finished: bool = False


class USBServer(Process):
  """This class is a server managing communication with USB devices through the
  :mod:`pysub` library.

  As :mod:`pyusb` is not process-safe in Python, running a server is the only
  option to allow multiple :class:`~crappy.blocks.Block` to use the library in
  parallel. This server simply sends the USB commands it receives to the USB
  devices, and returns back the answers. It features a quite complex
  architecture for managing the requests and properly starting up and shutting
  down.

  The server is a child of :obj:`multiprocessing.Process`.
  
  .. versionadded:: 1.5.2
  .. versionchanged:: 2.0.0 renamed from Usb_server to USBServer
  """

  initialized = False
  logger: Optional[logging.Logger] = None

  process: Optional[multiprocessing.context.Process] = None
  block_nr: int = 0
  devices: Dict[str, Device] = dict()

  # Objects for synchronizing with the server
  stop_event: Optional[multiprocessing.synchronize.Event] = None
  current_block: Optional[Synchronized] = None
  command_file: Optional[FileIO] = None
  answer_file: Optional[FileIO] = None
  shared_lock: Optional[multiprocessing.synchronize.RLock] = None
  block_dict: Dict[int, BlockObjects] = dict()

  def __init__(self,
               current_block: Synchronized,
               command_file: FileIO,
               answer_file: FileIO,
               block_dict: Dict[int, BlockObjects],
               stop_event: multiprocessing.synchronize.Event,
               log_queue: multiprocessing.queues.Queue,
               log_level: Optional[int]) -> None:
    """Sets the arguments.

    Args:
      current_block: A :obj:`multiprocessing.Value` storing the index of
        the :class:`~crappy.blocks.Block` currently allowed to communicate with
        the server.
      command_file: The handle to a file where the USB commands to send will be
        written.
      answer_file: The handle to a file where to write the answers from the USB
        devices.
      block_dict: A :obj:`dict` indicating for each index which
        :class:`~crappy.blocks.Block` it corresponds to.
      stop_event: A :obj:`multiprocessing.Event` indicating the server when it
        should stop running.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to
        the main :obj:`~logging.Logger`, only used in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
    """

    super().__init__(name=f'crappy.{type(self).__name__}')

    # Objects for synchronizing with the server
    self._current_block = current_block
    self._command_file = command_file
    self._answer_file = answer_file
    self._block_dict = block_dict
    self._stop_event = stop_event

    self._log_queue = log_queue
    self._logger: Optional[logging.Logger] = None
    self._log_level = log_level

    # Keeping a track of the number of connected blocks for each FT232H
    self._dev_count = dict()
    for ser_num in set(block.ser_num for block in self._block_dict.values()):
      self._dev_count[ser_num] = sum(
        1 for _ in (block for block in self._block_dict.values()
                    if block.ser_num == ser_num))

  @classmethod
  def register(cls, ser_num: Optional[str] = None) -> USBArgsType:
    """Allows a :class:`~crappy.blocks.Block` to register for communicating
    with the server. This Block is then given the necessary information for
    communication.

    Args:
      ser_num: The serial number of the FT232H to communicate with, as a
        :obj:`str`.

    Returns:
      A :obj:`tuple` containing the necessary information for other objects to
      communicate with the server. This information is for example given as
      arguments to :class:`~crappy.tool.ft232h.FT232HServer` objects.
    
    .. versionadded:: 2.0.0
    """

    # Initializing the synchronization objects
    if not cls.initialized:
      cls._initialize()

    # Assigning an index to the calling Block
    cls.block_nr += 1
    index = cls.block_nr

    # If there's only one device connected and the serial number is not
    # specified, using the connected device
    if len(cls.devices) == 1 and (ser_num is None or ser_num == ""):
      device = list(cls.devices.values())[0]

    # Otherwise, making sure the serial number is one of the connected devices
    else:
      try:
        device = cls.devices[ser_num]
      except KeyError:
        raise IOError(f"No FT232H detected with serial number {ser_num} !")

    # Storing the relevant attributes for the calling Block
    lock = RLock()
    cls.block_dict[index] = BlockObjects(ser_num=ser_num,
                                         lock=lock,
                                         device=device)

    return (index, lock, cls.command_file, cls.answer_file, cls.shared_lock,
            cls.current_block)

  @classmethod
  def start_server(cls,
                   log_queue: multiprocessing.queues.Queue,
                   log_level: int) -> None:
    """Initializes and starts the USB server Process.

    Args:
      log_queue: The :obj:`multiprocessing.Queue` carrying the log messages 
        from the server Process to Crappy's centralized log handler. Only used 
        in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.

    .. versionadded:: 2.0.0 *log_queue* and *log_level* arguments
    """

    cls.process = cls(current_block=cls.current_block,
                      command_file=cls.command_file,
                      answer_file=cls.answer_file,
                      block_dict=cls.block_dict,
                      stop_event=cls.stop_event,
                      log_queue=log_queue,
                      log_level=log_level)
    cls.process.start()

  @classmethod
  def stop_server(cls) -> None:
    """If the server was started, tries to stop it gently and if not successful
    terminates it.

    .. versionadded:: 2.0.0
    """

    if cls.process is not None:
      cls.stop_event.set()
      cls.log(logging.INFO, "Stop event set, waiting for the USB server to "
                            "finish")
      cls.process.join(0.2)

      if cls.process.is_alive():
        cls.log(logging.WARNING, "The USB server process did not stop "
                                 "correctly, killing it !")
        cls.process.terminate()

  @classmethod
  def _initialize(cls) -> None:
    """Sets the synchronization attributes and detects all the connected
    FT232H devices."""

    cls.devices = cls._get_devices()

    cls.stop_event = Event()
    cls.current_block = Value('B')
    cls.command_file = TemporaryFile(buffering=0)
    cls.answer_file = TemporaryFile(buffering=0)
    cls.shared_lock = RLock()

    cls.initialized = True

  @classmethod
  def log(cls, level: int, msg: str) -> None:
    """Wrapper for recording log messages.

    Also instantiates the :obj:`~logging.Logger` on the first message.

    Args:
      level: The logging level of the message, as an :obj:`int`.
      msg: The message to log, as a :obj:`str`.

    .. versionadded:: 2.0.0
    """

    if cls.logger is None:
      cls.logger = logging.getLogger(f'crappy.{cls.__name__}')

    cls.logger.log(level, msg)

  @staticmethod
  @contextmanager
  def acquire_timeout(lock: multiprocessing.synchronize.RLock,
                      timeout: float) -> bool:
    """Short context manager for acquiring a :obj:`multiprocessing.Lock` with a 
    specified timeout.

    Args:
      lock: The Lock to acquire.
      timeout: The timeout for acquiring the Lock, as a :obj:`float`.

    Returns:
      :obj:`True` if the Lock was successfully acquired, :obj:`False`
      otherwise.

    .. versionadded:: 2.0.0
    """

    ret = False
    try:
      ret = lock.acquire(timeout=timeout)
      yield ret
    finally:
      if ret:
        lock.release()

  @staticmethod
  def _get_devices() -> Dict[str, Any]:
    """Detects all the connected FT232H devices and returns them as a 
    :obj:`dict`.

    Returns:
      A :obj:`dict` containing as keys the detected serial numbers and as
      values the handles in the :mod:`pyusb` module to the associated FT232H
      devices.
    """

    # Searching for the FT232H devices
    devices: List[Device] = list(find(find_all=True,
                                      idVendor=0x0403,
                                      idProduct=0x6014))
    if not devices:
      raise IOError("No FT232H connected !")

    dev_dict = {}

    # Storing the found devices
    for device in devices:
      try:
        dev_dict[device.serial_number] = device

      except ValueError:
        # If there's only one FT232H connected, it can lack a serial number
        if len(devices) == 1:
          dev_dict[''] = devices[0]
        # Otherwise, every FT232H must have a serial number
        else:
          raise ValueError('Please set a serial number for each FT232H ! It '
                           'can be done using the dedicated crappy tool')

    return dev_dict

  def run(self) -> None:
    """The main loop of the server.

    Waits for a :class:`~crappy.blocks.Block` to acquire control, reads its 
    command, sends it to the correct USB device, reads the answer from the USB 
    device and sends it back to the Block in control. Then, does the same with 
    the next Block getting control over the server.

    .. versionadded:: 2.0.0
    """

    self._set_logger()
    self._log(logging.INFO, "Logger configured")

    # Disabling the KeyboardInterrupt exceptions, to avoid disruptions
    self._log(logging.WARNING, "Disabling KeyboardInterrupt for the server !")
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    while not self._stop_event.is_set():

      # Exiting if all the devices have been closed
      if all(dev.finished for dev in self._block_dict.values()):
        self._log(logging.INFO, "Server finished after all devices were "
                                "closed")
        break

      # A Block has acquired the lock and shared its index
      if self._current_block.value:
        # Reading the index and resetting the value
        index = self._current_block.value
        self._current_block.value = 0
        self._log(logging.DEBUG, f"Block with index {index} now has control")

        # The Block has finished writing the command
        with self.acquire_timeout(self._block_dict[index].lock, 1) as acquired:
          if acquired:
            self._log(logging.DEBUG, f"Acquired lock of Block with index "
                                     f"{index}")
            # Reading the command
            self._command_file.seek(0)
            command = self._command_file.read()
            # Resetting the command file
            self._command_file.seek(0)
            self._command_file.truncate(0)

            self._log(logging.DEBUG, f"Received command {command}")

            try:
              # Sends the command to the device and returns the answer
              answer = self._send_command(command,
                                          self._block_dict[index].device,
                                          self._block_dict[index].ser_num,
                                          index)
            except (USBTimeoutError, TimeoutError):
              # Double-checking the timeout error
              answer = self._send_command(command,
                                          self._block_dict[index].device,
                                          self._block_dict[index].ser_num,
                                          index)

            self._log(logging.DEBUG,
                      f"Got answer {answer} from device with serial number "
                      f"{self._block_dict[index].ser_num}")

            # Resetting the answer file
            self._answer_file.seek(0)
            self._answer_file.truncate(0)
            # Writing the answer in the answer file
            try:
              self._answer_file.write(answer)
            except TypeError:
              # Sometimes for an unknown reason the answer is None
              raise IOError("Got an unexpected USB answer from an FT232H !")

          else:
            raise TimeoutError(
              f"Could not acquire the Block lock of Block with index {index} "
              f"driving the FT232H with serial number "
              f"{self._block_dict[index].ser_num} within 1s, aborting !")

    if self._stop_event.is_set():
      self._log(logging.INFO, "Server finished after stop event was set")

  def _send_command(self,
                    command: bytes,
                    device: Device,
                    serial_nr: str,
                    index: int) -> bytes:
    """Sends commands to a USB device and returns the answer.

    Args:
      command: The command to send to the device. The bytes are arranged in a
        specific way for each type of command.
      device: The :mod:`pyusb` Device to which the commands are sent.
      serial_nr: The serial number of the :mod:`pyusb` device.
      index: The index of the Block currently controlling the server.

    Returns:
      The index of the command, followed by the answer from the USB device if
      applicable.
    """

    command = command.split(b',')

    # Control transfer out
    if command[0] == b'00':
      return b','.join((b'00',
                        str(device.ctrl_transfer(int(command[1]),
                                                 int(command[2]),
                                                 int(command[3]),
                                                 int(command[4]),
                                                 command[5],
                                                 int(command[6]))).encode()))
    # Control transfer in
    elif command[0] == b'01':
      return b','.join((b'01',
                        bytes(device.ctrl_transfer(*[int(arg) for arg in
                                                     command[1:]]))))
    # Write operation
    elif command[0] == b'02':
      return b','.join((b'02',
                        str(device.write(int(command[1]), command[2],
                                         int(command[3]))).encode()))
    # Read operation
    elif command[0] == b'03':
      return b','.join((b'03',
                        bytes(device.read(*[int(arg) for arg
                                            in command[1:]]))))
    # Checks whether the kernel driver is active
    # It doesn't actually interact with the device
    elif command[0] == b'04':
      return b','.join((b'04',
                        b'1' if device.is_kernel_driver_active(int(command[1]))
                        else b'0'))
    # Detaches the kernel driver
    # It doesn't actually interact with the device
    elif command[0] == b'05':
      device.detach_kernel_driver(int(command[1]))
      return b'05,'
    # Sets the device configuration
    elif command[0] == b'06':
      device.set_configuration()
      return b'06,'
    # Custom command getting information from the current configuration
    elif command[0] == b'07':
      info = self._return_config_info(device)
      return b','.join((b'07',
                        str(info[0]).encode(), str(info[1]).encode(),
                        str(info[2]).encode(), str(info[3]).encode()))
    # When a block is leaving, if it's the last one associated with a given
    # ft232h then it should release the internal resources
    # It doesn't actually interact with the device
    elif command[0] == b'08':
      self._dev_count[serial_nr] -= 1
      return b','.join((b'08',
                        b'1' if not self._dev_count[serial_nr] else b'0'))

    # Checks whether the internal resources have been released or not
    # It doesn't actually interact with the device
    elif command[0] == b'09':
      return b','.join((b'09', b'1' if device._ctx.handle else b'0'))
    # Releases the USB interface
    # It doesn't actually interact with the device
    elif command[0] == b'10':
      util.release_interface(device, int(command[1]))
      return b'10,'
    # Detaches the kernel driver
    # It doesn't actually interact with the device
    elif command[0] == b'11':
      device.attach_kernel_driver(int(command[1]))
      return b'11,'
    # Releases all the resources used by :mod:`pyusb` for a given device
    # It doesn't actually interact with the device
    elif command[0] == b'12':
      util.dispose_resources(device)
      return b'12,'
    # Registers a block as gone
    # It doesn't actually interact with the device
    elif command[0] == b'13':
      self._block_dict[index].finished = True
      return b'13,'

  def _set_logger(self) -> None:
    """Instantiates and sets up the :obj:`~logging.Logger` for recording log
    messages."""

    logger = logging.getLogger(self.name)

    # Disabling logging if requested
    if self._log_level is not None:
      logger.setLevel(self._log_level)
    else:
      logging.disable()

    # On Windows, the messages need to be sent through a Queue for logging
    if get_start_method() == "spawn" and self._log_level is not None:
      queue_handler = logging.handlers.QueueHandler(self._log_queue)
      queue_handler.setLevel(self._log_level)
      logger.addHandler(queue_handler)

    self._logger = logger

  def _log(self, level: int, msg: str) -> None:
    """Wrapper for recording log messages.

    Args:
      level: The logging level of the message, as an :obj:`int`.
      msg: The message to lof, as a :obj:`str`.
    """

    if self._logger is None:
      return
    self._logger.log(level, msg)

  @staticmethod
  def _return_config_info(device) -> Tuple[int, int, int, int]:
    """Returns some configuration information from a USB object.

    Args:
      device: A :obj:`usb.core.Device`

    Returns:
      The index, in endpoint, out endpoint and maximum packet size of a USB
      device.
    """

    interface = device.get_active_configuration()[(0, 0)]
    index = interface.bInterfaceNumber + 1
    in_ep, out_ep = sorted([ep.bEndpointAddress for ep in interface])[:2]
    max_packet_size = interface[0].wMaxPacketSize
    return index, in_ep, out_ep, max_packet_size



# coding: utf-8

from typing import Optional
import logging
from  warnings import warn

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  from serial import Serial
except (ModuleNotFoundError, ImportError):
  Serial = OptionalModule("pyserial")


class OrientalARDK(Actuator):
  """This class can drive an Oriental Motor's ARD-K stepper motor driver in
  speed or in position.

  It communicates with the stepper motor over a serial connection. This class
  was designed so that the :class:`~crappy.blocks.Machine` Block drives several
  of its instances at a time, corresponding to different axes to drive.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Oriental to OrientalARDK
  """

  def __init__(self,
               baudrate: int = 115200,
               port: str = '/dev/ttyUSB0',
               gain: float = 1/.07) -> None:
    """Sets the instance attributes and initializes the parent class.

    Args:
      baudrate: The baudrate to use for the serial communication.
      port: The path to the serial port to use for communication.
      gain: The gain to apply to speed commands, in `mm/min`. The default value
        corresponds to `0.07mm/min` for a command value of `1`.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._ser = None

    super().__init__()

    self._baudrate = baudrate
    self._port = port
    self._gain = gain

    self._prev_set_speed = 0

  def open(self) -> None:
    """Opens the serial connection to the motor and initializes the motor."""

    # Opening the serial connection to the actuator
    self.log(logging.INFO, f"Opening the serial port {self._port} with "
                           f"baudrate {self._baudrate}")
    self._ser = Serial(self._port, baudrate=self._baudrate, timeout=0.1)

    # Checking which of the four motors is the one connected to the chosen port
    for i in range(1, 5):
      self.log(logging.DEBUG, f"Writing b'TALK{i}\\n' to port {self._port}")
      self._ser.write(f"TALK{i}\n".encode())
      ret = self._ser.readlines()
      self.log(logging.DEBUG, f"Read {ret} on port {self._port}")

      # Displaying the connected motor
      if f"{i}>".encode() in ret:
        motors = ['A', 'B', 'C', 'D']
        self.log(logging.INFO, f"Motor connected to port {self._port} is "
                               f"{motors[i-1]}")
        break

    self._clear_errors()

    # Setting the acceleration and deceleration
    self.log(logging.DEBUG, f"Writing b'TA .1\\n' to port {self._port}")
    self._ser.write(b"TA .1\n")
    self.log(logging.DEBUG, f"Writing b'TD .1\\n' to port {self._port}")
    self._ser.write(b"TD .1\n")

  def set_speed(self, cmd: float) -> None:
    """Sets the target speed for the motor.

    Also manages the sign of the speed, i.e. the direction of the movement.
    Features a check to avoid sending the same speed command multiple times.

    Args:
      cmd: The target speed value, no units. The actual speed reached by the
        motor in `mm/min` depends on the gain that was set.
    """

    # Applying the gain and clamping the value to match the limits
    speed = min(100, int(abs(cmd * self._gain) + .5))

    # A speed of zero means a stop
    if speed == 0:
      self.stop()
      return

    # Taking the sign into consideration
    sign = int(cmd / abs(cmd))
    signed_speed = sign * speed

    # If the value is the same as the previous, do nothing
    if signed_speed == self._prev_set_speed:
      return

    # Stopping the motor if the direction changed
    dir_chg = self._prev_set_speed * sign < 0
    if dir_chg:
      self.stop()

    # Writing the target speed value
    self.log(logging.DEBUG, f"Writing b'VR {abs(speed)}\\n' to port "
                            f"{self._port}")
    self._ser.write(f'VR {abs(speed)}\n'.encode())
    # Writing the target direction
    if sign > 0:
      self.log(logging.DEBUG, f"Writing b'MCP\\n' to port {self._port}")
      self._ser.write(b"MCP\n")
    else:
      self.log(logging.DEBUG, f"Writing b'MCN\\n' to port {self._port}")
      self._ser.write(b"MCN\n")

    # Storing the written value
    self._prev_set_speed = signed_speed

  def set_position(self, position: float, speed: Optional[float]) -> None:
    """Sets the target position for the motor.

    Args:
      position: The target position to reach, in arbitrary units.
      speed: The speed to use for reaching the target position, in arbitrary
        units. A speed must be given, otherwise an exception is raised.
    
    .. versionchanged:: 2.0.0 *speed* is now a mandatory argument
    """

    if speed is None:
      raise ValueError("The OrientalARDK actuator needs both a position and a "
                       "speed command when driven in position mode !")

    self.log(logging.DEBUG, f"Writing b'VR {abs(speed)}' to port {self._port}")
    self._ser.write(f'VR {abs(speed)}'.encode())
    self.log(logging.DEBUG, f"Writing b'MA {position}' to port {self._port}")
    self._ser.write(f'MA {position}'.encode())

  def get_position(self) -> float:
    """Reads and returns the current position of the motor.

    .. versionchanged:: 1.5.2 renamed from get_pos to get_position
    """

    # Sending the read command
    self._ser.flushInput()
    self.log(logging.DEBUG, f"Writing b'PC\\n' to port {self._port}")
    self._ser.write(b'PC\n')
    self._ser.readline()

    # Reading the position and returning it
    actuator_pos = str(self._ser.readline())
    self.log(logging.DEBUG, f"Read {actuator_pos} on port {self._port}")
    try:
      return float(actuator_pos[4:-3])
    except ValueError:
      return 0

  def stop(self) -> None:
    """Sends a command for stopping the motor."""

    if self._ser is not None:
      self.log(logging.DEBUG, f"Writing b'SSTOP\\n' to port {self._port}")
      self._ser.write(b"SSTOP\n")

  def close(self) -> None:
    """Closes the serial connection to the motor."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()

  def _clear_errors(self) -> None:
    """Sends a command for clearing any serial error from the motor."""

    self.log(logging.DEBUG, f"Writing b'ALMCLR\\n' to port {self._port}")
    self._ser.write(b"ALMCLR\n")



# coding: utf-8

from machine import reset
from select import poll, POLLIN
from sys import stdin, stdout
from utime import sleep_ms, ticks_ms
from struct import pack

"""To use this template, replace the lines between the >>> and <<< by your own 
code. See the examples/blocks/ucontroller folder for a running example.
Here, this example simply sets a GPIO low and high at a given frequency and 
sends back the number of cycles to the PC every 10 cycles. Connect a LED to the
GPIO to see the result !
"""

# Creating a poll object
p = poll()
p.register(stdin, POLLIN)
msg = ''


def read():
  """Method for reading the labels coming from the PC at the beginning of the
  program."""

  i = 0
  while True:
    val = stdin.readline().strip()
    if val:
      return val
    sleep_ms(50)
    if i > 9:
      break
    i += 1


"""Enters an infinite loop and exits only upon reception of 'goXY'. X is the
number of command labels, Y is the number of labels. The labels are then 
received and stored. This setup prevents the program from doing anything before 
it is told to."""
while True:
  if p.poll(0):
    msg = stdin.readline().strip()
    # Program has been launched on PC
    if msg.startswith('go'):
      commands = {}
      # Getting the commands
      for _ in range(int(msg[2])):
        value = read()
        commands[int(value[0])] = str(value[1:])
      labels = {}
      # Getting the labels
      for _ in range(int(msg[3])):
        value = read()
        labels[str(value[1:])] = int(value[0])
      # If 't(s)' in labels, the program should send back timestamps
      send_t = 't(s)' in labels
      msg = ''
      break
  sleep_ms(50)

# >>>>>>>>

# Here we initialize the variables we want to use
from machine import Pin
freq = 1
count = 0
pin = Pin(13, Pin.OUT)
pin.off()

# <<<<<<<<


def send_to_pc(var, label):
  """Sends back data to the PC.

  If send_t is True, also sends back the timestamp in milliseconds. The time
  is encoded as an integer, the data as a float. Inbetween, a signed char
  indicates the index of the label that is being sent. This index has been set
  by the PC and sent in the labels dict.

  See below for a use case.
  """

  try:
    if send_t:
      stdout.buffer.write(pack('<ibf', ticks_ms(), labels[label], var))
    else:
      stdout.buffer.write(pack('<bf', labels[label], var))
  except KeyError:
    pass


"""The main loop of this script.
It reads the incoming messages, and does a user-defined action when no message
is waiting."""
while True:
  # Reads the incoming messages if any is waiting
  if p.poll(0):
    msg = stdin.readline().strip()

  # Upon reception of 'stop!', the program ends
  if msg == 'stop!':
    break

  # Acquiring commands
  # Their names are those of the cmd_labels in the UController block
  # In the example only freq can be updated as it is the only element of
  # cmd_labels
  if msg:
    globals()[commands[int(msg[0])]] = float(msg[1:])
    msg = ''

  while not p.poll(0):
    # Here should be the main task of your script
    # It will be called repeatedly, but the loop will be interrupted upon
    # reception of a command

    # >>>>>>>>

    # Blinking the GPIO
    pin.on()
    sleep_ms(int(500 / freq))
    pin.off()
    sleep_ms(int(500 / freq))

    # Sending back the number of cycles (count) under the label 'nr'
    # Only the labels present in the labels argument of the UController block
    # can be sent back
    # Make sure that your Crappy and MicroPython scripts use the same naming
    if not count % 10:
      # Example of a call to send_to_pc
      # First argument is the data to send, second is its label
      # Also sends back the associated timestamp if send_t is True
      send_to_pc(count, 'nr')

    count += 1

    # <<<<<<<<

"""Reset the microcontroller, so that it is then stuck again in the first 
infinite loop and waiting for the 'go' message."""
reset()



# coding: utf-8

from typing import Union, Dict
import logging

from .meta_path import Path, ConditionType


class Conditional(Path):
  """This Path returns one of three possible output values, based on two given
  conditions.

  It is especially useful for controlling processes that need to behave
  differently based on input values, e.g. for preventing a heating element
  from overheating, or a motor from driving too far.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               condition1: Union[str, ConditionType],
               condition2: Union[str, ConditionType],
               value1: float,
               value2: float,
               value0: float = 0) -> None:
    """Sets the args and initializes the parent class.

    Args:
      condition1: The first condition checked by the Path. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      condition2: The second condition checked by the path. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      value1: Value to send when ``condition1`` is met.
      value2: Value to send when ``condition2`` is met and ``condition1`` is
        not met.
      value0: Value to send when neither ``condition1`` nor ``condition2`` are
        met.

    Note:
      This Generator Path never ends, it doesn't have a stop condition.

    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionremoved:: 1.5.10 *verbose* argument
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    .. versionchanged:: 2.0.0 renamed from Protection to Conditional
    """

    super().__init__()

    # Setting the attributes
    self._value0 = value0
    self._value1 = value1
    self._value2 = value2
    self._condition1 = self.parse_condition(condition1)
    self._condition2 = self.parse_condition(condition2)
    self._prev = self._value0

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Sends either ``value1`` if ``condition1`` is met, or ``value2`` if only
    ``condition2`` is met, or ``value0`` if none of the conditions are met."""

    # Case when data has been received
    if any(data.values()):

      # Send value1 if the first condition is met
      if self._condition1(data):
        self.log(logging.DEBUG, "Condition 1 met")
        self._prev = self._value1
        return self._value1

      # Send value2 if only the second condition is met
      elif self._condition2(data):
        self.log(logging.DEBUG, "Condition 2 met")
        self._prev = self._value2
        return self._value2

      # Send value0 if no condition is met
      else:
        self.log(logging.DEBUG, "Neither condition 1 nor condition 2 met")
        self._prev = self._value0
        return self._value0

    # If no data received, return the last sent value
    else:
      return self._prev



# coding: utf-8

from time import time, strftime, gmtime
from typing import Optional, Tuple, Dict, Any
import numpy as np
import logging
from warnings import warn

from .meta_camera import Camera
from .._global import OptionalModule

try:
  from ximea import xiapi
except (ModuleNotFoundError, ImportError):
  xiapi = OptionalModule("ximea", "To use XiAPI cameras, please install the "
                         "official ximea Python module")

# Camera models that were added in Crappy
SUPPORTED = ('MQ042MG-CM', 'MC124MG-SY-UB')

# Available data formats for the supported cameras
DATA_FORMATS = {'MQ042MG-CM': {'XI_MONO8': 'Mono (8 bits)',
                               'XI_MONO16': 'Mono (16 bits)',
                               'XI_RAW8': 'Raw (8 bits)',
                               'XI_RAW16': 'Raw (16 bits)'},
                'MC124MG-SY-UB': {'XI_MONO8': 'Mono (8 bits)',
                                  'XI_MONO16': 'Mono (16 bits)',
                                  'XI_RAW8': 'Raw (8 bits)',
                                  'XI_RAW16': 'Raw (16 bits)'}
                }
DATA_FORMATS_INV = {num: dict(zip(dic.values(), dic.keys()))
                    for num, dic in DATA_FORMATS.items()}

# Available framerate modes for the supported cameras
FRAMERATE_MODES = {'MQ042MG-CM': 
                   {'XI_ACQ_TIMING_MODE_FREE_RUN': 'Free run',
                    'XI_ACQ_TIMING_MODE_FRAME_RATE': 'Framerate target'},
                   'MC124MG-SY-UB':
                   {'XI_ACQ_TIMING_MODE_FREE_RUN': 'Free run',
                    'XI_ACQ_TIMING_MODE_FRAME_RATE_LIMIT': 'Framerate limit'}
                   }
FRAMERATE_MODES_INV = {num: dict(zip(dic.values(), dic.keys()))
                       for num, dic in FRAMERATE_MODES.items()}

# Available downsampling modes for the supported cameras
DOWNSAMPLING_MODES = {'MC124MG-SY-UB': {'XI_DWN_1x1': '1x1',
                                        'XI_DWN_2x2': '2x2'}}
DOWNSAMPLING_MODES_INV = {num: dict(zip(dic.values(), dic.keys()))
                          for num, dic in DOWNSAMPLING_MODES.items()}


class XiAPI(Camera):
  """This class can read images from any of the Ximea cameras.

  It heavily relies on the :mod:`ximea` module, distributed by Ximea, which is
  itself a wrapper around the XiAPI low-level library.

  Note:
    Both the Python module and the camera drivers have to be installed from
    Ximea's website in order for this class to run.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Xiapi to XiAPI
  """

  def __init__(self) -> None:
    """Instantiates a Ximea Camera and a Ximea Image object."""

    super().__init__()

    self._cam = None
    self._started: bool = False
    self._model: Optional[str] = None
    self._timeout: Optional[int] = None

    self._cam = xiapi.Camera()
    self._img = xiapi.Image()

    # Stores the last requested or read trigger mode value
    self._trig = 'Free run'

  def open(self,
           serial_number: Optional[str] = None,
           timeout: Optional[int] = None,
           trigger: Optional[str] = None,
           data_format: Optional[str] = None,
           exposure_time_us: Optional[int] = None,
           gain: Optional[float] = None,
           auto_exposure_auto_gain: Optional[bool] = None,
           gamma_y: Optional[float] = None,
           gamma_c: Optional[float] = None,
           sharpness: Optional[float] = None,
           image_width: Optional[int] = None,
           image_height: Optional[int] = None,
           x_offset: Optional[int] = None,
           y_offset: Optional[int] = None,
           framerate_mode: Optional[str] = None,
           framerate: Optional[float] = None,
           downsampling_mode: Optional[str] = None,
           width: Optional[int] = None,
           height: Optional[int] = None,
           xoffset: Optional[int] = None,
           yoffset: Optional[int] = None,
           exposure: Optional[float] = None,
           AEAG: Optional[bool] = None,
           external_trig: Optional[bool] = None,) -> None:
    """Opens the connection to the camera, instantiates the available settings
    and starts the acquisition.

    Also sets custom values for the settings if provided by the user, otherwise
    sets them to their default.

    Args:
      serial_number: A :obj:`str` containing the serial number of the camera to
        open, in case several cameras are connected. If not provided and
        several cameras are available, one of them will be opened randomly.

        .. versionchanged:: 2.0.0 renamed from *sn* to *serial_number*

      timeout: The number of milliseconds the camera is allowed to wait for an
        image before raising a :exc:`TimeoutError`, as an :obj:`int`. Mostly
        useful when using an external trigger, or a very long exposure time.
        The default is 5000ms.

        .. versionadded:: 2.0.2

      trigger: Drives the trigger status of the camera. Can be either
        `'Free run'`, `'Hdw after config'`, or `'Hardware'`. The default is
        `'Free run'`.

        .. versionadded:: 2.0.0

      data_format: The data format to use for acquisition as a :obj:`str` (e.g.
        `'Mono (8 bits)'`). The available formats depend on the model of the
        camera. This setting is only implemented for a subset of all the Ximea
        cameras, get in touch with the maintainers to implement in on other
        models.

        .. versionadded:: 2.0.2

      exposure_time_us: The exposure time to use for acquiring images, in µs.
        Only has an effect if ``auto_exposure_auto_gain`` is set to
        :obj:`False`, and if ``framerate_mode`` is not in `'Free run'`.
        Depending on the camera might set the *target* or the *limit*
        framerate. The range of values depends on the model of the driven
        camera. This setting is only implemented for a subset of all the Ximea
        cameras, get in touch with the maintainers to implement in on other
        models.

        .. versionadded:: 2.0.2

      gain: The gain in dB to apply to the acquired images, as a :obj:`float`.
        Only has an effect if ``auto_exposure_auto_gain`` is set to
        :obj:`False`. The valid range of values depends on the model of the
        driven camera.
      auto_exposure_auto_gain: When set to :obj:`True`, overwrites the
        ``exposure_time_us`` and ``gain`` values and drives the exposure and
        the gain automatically.

        .. versionadded:: 2.0.2

      gamma_y: The luminosity gamma value, for color images only, as a
        :obj:`float`. The valid range of values depends on the model of the
        driven camera.

        .. versionadded:: 2.0.2

      gamma_c: The chromaticity gamma value, for color images only, as a
        :obj:`float`. The valid range of values depends on the model of the
        driven camera.

        .. versionadded:: 2.0.2

      sharpness: The sharpness strength value, for color images only, as a
        :obj:`float`. The valid range of values depends on the model of the
        driven camera.
        
        .. versionadded:: 2.0.2

      image_width: The width of the ROI to acquire in pixels, as an :obj:`int`.
        The valid range of values depends on the model of the driven camera, 
        and on the ``downsampling_mode`` if supported.
        
        .. versionadded:: 2.0.2
        
      image_height: The height of the ROI to acquire in pixels, as an 
        :obj:`int`. The valid range of values depends on the model of the 
        driven camera, and on the ``downsampling_mode`` if supported.
        
        .. versionadded:: 2.0.2
        
      x_offset: The X offset of the ROI to acquire in pixels, as an :obj:`int`. 
        The valid range of values depends on the model of the driven camera, 
        on the selected ``width``, and on the ``downsampling_mode`` if 
        supported.
        
        .. versionadded:: 2.0.2
        
      y_offset: The Y offset of the ROI to acquire in pixels, as an :obj:`int`. 
        The valid range of values depends on the model of the driven camera, 
        on the selected ``height``, and on the ``downsampling_mode`` if 
        supported.
        
        .. versionadded:: 2.0.2
        
      framerate_mode: The mode for driving the acquisition framerate. Can be 
        one of `'Free run'`, `'Framerate target'` or `'Framerate limit'`, 
        but not all camera models support the three options. This setting is 
        only implemented for a subset of all the Ximea cameras, get in touch 
        with the maintainers to implement in on other models.
        
        .. versionadded:: 2.0.2
        
      framerate: Either the target or the limit framerate for image 
        acquisition, depending on the ``framerate_mode`` value, as a 
        :obj:`float`. The valid range of values depends on the model of the 
        driven camera, and on the ``exposure_time_us`` value. Only has an 
        effect if ``framerate_mode`` is not set to `'Free run'`. This setting 
        is only implemented for a subset of all the Ximea cameras, get in touch 
        with the maintainers to implement in on other models.
        
        .. versionadded:: 2.0.2
        
      downsampling_mode: The downsampling mode to apply to the acquired images,
        as a :obj:`str` (e.g. `'2x2'`). This setting is only implemented for a 
        subset of all the Ximea cameras, get in touch with the maintainers to 
        implement in on other models.
        
        .. versionadded:: 2.0.2
        
      width: .. deprecated:: 2.0.2 Use ``image_width`` instead.
      height: .. deprecated:: 2.0.2 Use ``image_height`` instead.
      xoffset: .. deprecated:: 2.0.2 Use ``x_offset`` instead.
      yoffset: .. deprecated:: 2.0.2 Use ``y_offset`` instead.
      exposure: .. deprecated:: 2.0.2 Use ``exposure_time_us`` instead.
      AEAG: .. deprecated:: 2.0.2 Use ``auto_exposure_auto_gain`` instead.
      external_trig: .. deprecated:: 2.0.0 Use ``trigger`` instead.
    """

    # Managing the deprecated arguments
    if width is not None:
      warn("The width argument is deprecated for the XiAPI camera, use "
           "image_width instead !", FutureWarning)
      image_width = width
    if height is not None:
      warn("The height argument is deprecated for the XiAPI camera, use "
           "image_height instead !", FutureWarning)
      image_height = height
    if xoffset is not None:
      warn("The xoffset argument is deprecated for the XiAPI camera, use "
           "x_offset instead !", FutureWarning)
      x_offset = xoffset
    if yoffset is not None:
      warn("The yoffset argument is deprecated for the XiAPI camera, use "
           "y_offset instead !", FutureWarning)
      y_offset = yoffset
    if exposure is not None:
      warn("The exposure argument is deprecated for the XiAPI camera, use "
           "exposure_time_us instead !", FutureWarning)
      exposure_time_us = exposure
    if AEAG is not None:
      warn("The AEAG argument is deprecated for the XiAPI camera, use "
           "auto_exposure_auto_gain instead !", FutureWarning)
      auto_exposure_auto_gain = AEAG
    if external_trig is not None:
      warn("The external_trig argument is deprecated for the XiAPI camera, use"
           " trigger instead !", FutureWarning)
      trigger = 'Hdw after config'

    self._timeout = timeout

    # First, checking if there are connected cameras
    num_dev = self._cam.get_number_devices()
    if not num_dev:
      raise IOError("No connected Ximea camera was detected !")
    else:
      self.log(logging.INFO, f"Detected {num_dev} connected Ximea camera(s)")

    # Opening the camera by serial number if any was provided
    if serial_number is not None:
      self.log(logging.INFO, f"Opening the connection to the camera with "
                             f"serial number {serial_number}")
      self._cam.open_device_by_SN(serial_number)
    else:
      self.log(logging.INFO, "Opening the connection to a default camera")
      self._cam.open_device()

    # Checking the model of the opened camera and logging to the user
    self._model = self._cam.get_device_name().decode()
    self.log(logging.INFO, f"Opened the Ximea camera model "
                           f"{self._model}, with serial number "
                           f"{self._cam.get_device_sn().decode()}")
    if self._model not in SUPPORTED:
      self.log(logging.WARNING, f"The model {self._model} was never "
                                f"specifically implemented in Crappy, some if "
                                f"its features might not be available ! Get "
                                f"in touch with the maintainers to have it "
                                f"implemented.")

    # Data format parameter
    if self._model in DATA_FORMATS:
      self.add_choice_setting('data_format',
                              DATA_FORMATS[self._model].values(),
                              self._get_data_format,
                              self._set_data_format,
                              self._get_data_format())

    # Gain and exposure parameters
    self.add_scale_setting('exposure_time_us',
                           self._cam.get_exposure_minimum(),
                           min(self._cam.get_exposure_maximum(), 100000),
                           self._get_exp,
                           self._set_exp,
                           10000)
    self.add_scale_setting('gain',
                           self._cam.get_gain_minimum(),
                           self._cam.get_gain_maximum(),
                           self._get_gain,
                           self._set_gain,
                           0.0)
    self.add_bool_setting('auto_exposure_auto_gain',
                          self._get_aeag,
                          self._set_aeag,
                          False)

    # Color-related parameters
    self.add_scale_setting('gamma_y',
                           self._cam.get_gammaY_minimum(),
                           self._cam.get_gammaY_maximum(),
                           self._get_gamma_y,
                           self._set_gamma_y,
                           self._cam.get_gammaY(),
                           self._cam.get_gammaY_increment())
    self.add_scale_setting('gamma_c',
                           self._cam.get_gammaC_minimum(),
                           self._cam.get_gammaC_maximum(),
                           self._get_gamma_c,
                           self._set_gamma_c,
                           self._cam.get_gammaC(),
                           self._cam.get_gammaC_increment())
    self.add_scale_setting('sharpness',
                           self._cam.get_sharpness_minimum(),
                           self._cam.get_sharpness_maximum(),
                           self._get_sharpness,
                           self._set_sharpness,
                           self._cam.get_sharpness(),
                           self._cam.get_sharpness_increment())

    # ROI parameters
    self.add_scale_setting('image_width',
                           self._cam.get_width_minimum(),
                           self._cam.get_width_maximum(),
                           self._get_w,
                           self._set_w,
                           self._get_w(),
                           self._cam.get_width_increment())
    self.add_scale_setting('image_height',
                           self._cam.get_height_minimum(),
                           self._cam.get_height_maximum(),
                           self._get_h,
                           self._set_h,
                           self._get_h(),
                           self._cam.get_height_increment())
    self.add_scale_setting('x_offset',
                           self._cam.get_offsetX_minimum(),
                           self._cam.get_offsetX_maximum(),
                           self._get_ox,
                           self._set_ox,
                           self._get_ox(),
                           self._cam.get_offsetX_increment())
    self.add_scale_setting('y_offset',
                           self._cam.get_offsetY_minimum(),
                           self._cam.get_offsetY_maximum(),
                           self._get_oy,
                           self._set_oy,
                           self._get_oy(),
                           self._cam.get_offsetY_increment())

    # External trigger parameter
    self.add_trigger_setting(self._get_extt, self._set_ext_trig)

    # Framerate parameters
    if self._model in FRAMERATE_MODES:
      self.add_choice_setting('framerate_mode',
                              FRAMERATE_MODES[self._model].values(),
                              self._get_framerate_mode,
                              self._set_framerate_mode,
                              self._get_framerate_mode())
      self.add_scale_setting('framerate',
                             self._cam.get_framerate_minimum(),
                             min(self._cam.get_framerate_maximum(), 500),
                             self._get_framerate,
                             self._set_framerate,
                             self._cam.get_framerate(),
                             self._cam.get_framerate_increment())

    # Downsampling parameter
    if self._model in DOWNSAMPLING_MODES:
      self.add_choice_setting('downsampling_mode',
                              DOWNSAMPLING_MODES[self._model].values(),
                              self._get_downsampling_mode,
                              self._set_downsampling_mode,
                              self._get_downsampling_mode())

    # Collecting the kwargs to set and setting them
    to_set = {name: arg for name, arg in zip(
        ('exposure_time_us', 'gain', 'auto_exposure_auto_gain', 'gamma_y',
         'gamma_c', 'sharpness', 'image_width', 'image_height', 'x_offset',
         'y_offset', 'framerate_mode', 'framerate', 'downsampling_mode',
         'trigger'),
        (exposure_time_us, gain, auto_exposure_auto_gain, gamma_y, gamma_c,
         sharpness, image_width, image_height, x_offset, y_offset,
         framerate_mode, framerate, downsampling_mode, trigger))
              if arg is not None}
    self.set_all(**to_set)

    # Starting the acquisition
    self.log(logging.INFO, "Starting the image acquisition")
    self._cam.start_acquisition()
    self._started = True

  def get_image(self) -> Tuple[Dict[str, Any], np.ndarray]:
    """Reads a frame from the camera, and returns it along with its metadata.

    The acquired metadata contains the following fields :

    * `'t(s)'`: The current timestamp as returned by :obj:`time.time`.
    * `'DateTimeOriginal'`: The current date up to the second as a valid exif
      tag, based on the value of `'t(s)'`.
    * `'SubsecTimeOriginal'`: The sub-second part of the current date as a
      valid exif tag, based on the value of `'t(s)'`.
    * `'XimeaSec'`: The number of seconds the camera has been up.
    * `'XimeaUSec'`: The decimal part of the above field, value in µs.
    * `'ImageWidth'`: The width of the acquired image, in pixels.
    * `'ImageHeight'`: The height of the acquired image, in pixels.
    * `'ExposureTime'`: The exposure time of the acquired image, in µs.
    * `'AbsoluteOffsetX'`: The offset of the acquired ROI along the X axis, in
      pixels.
    * `'AbsoluteOffsetY'`: The offset of the acquired ROI along the Y axis, in
      pixels.
    * `'DownsamplingX'`: The downsampling factor along the X axis.
    * `'DownsamplingY'`: The downsampling factor along the Y axis.
    * `'ImageUniqueID'`: The index of the acquired image, as returned by the
      camera.

    """

    if self._timeout is not None:
      self._cam.get_image(self._img, timeout=self._timeout)
    else:
      self._cam.get_image(self._img)

    # Returning metadata from the camera long with the captured image
    t = time()
    metadata = {'t(s)': t,
                'DateTimeOriginal': strftime("%Y:%m:%d %H:%M:%S", gmtime(t)),
                'SubsecTimeOriginal': f'{t % 1:.6f}',
                'XimeaSec': self._img.tsSec,
                'XimeaUSec': self._img.tsUSec,
                'ImageWidth': self._img.width,
                'ImageHeight': self._img.height,
                'ExposureTime': self._img.exposure_time_us,
                'AbsoluteOffsetX': self._img.AbsoluteOffsetX,
                'AbsoluteOffsetY': self._img.AbsoluteOffsetY,
                'DownsamplingX': self._img.DownsamplingX,
                'DownsamplingY': self._img.DownsamplingY,
                'ImageUniqueID': self._img.acq_nframe}

    return metadata, self._img.get_image_data_numpy()

  def close(self) -> None:
    """Closes the connection to the camera and releases the resources."""

    if self._cam is not None:
      self.log(logging.INFO, "Closing the connection to the camera")
      self._cam.close_device()

  def _get_data_format(self) -> str:
    """Returns the current data format of the acquired images."""

    return DATA_FORMATS[self._model][self._cam.get_imgdataformat()]

  def _get_exp(self) -> float:
    """Returns the exposure time, in microseconds."""

    return self._cam.get_exposure()

  def _get_aeag(self) -> bool:
    """Return the status of the Auto Exposure / Auto Gain setting."""

    return self._cam.is_aeag()

  def _get_gain(self) -> float:
    """Returns the gain, in dB."""

    return self._cam.get_gain()

  def _get_gamma_y(self) -> float:
    """Returns the current Gamma Y value of the camera."""

    return self._cam.get_gammaY()

  def _get_gamma_c(self) -> float:
    """Returns the current Gamma C value of the camera."""

    return self._cam.get_gammaC()
  
  def _get_sharpness(self) -> float:
    """Returns the current sharpness value of the camera."""

    return self._cam.get_sharpness()

  def _get_w(self) -> int:
    """Returns the width in pixels for selecting a region of interest."""

    return self._cam.get_width()

  def _get_h(self) -> int:
    """Returns the height in pixels for selecting a region of interest."""

    return self._cam.get_height()

  def _get_ox(self) -> int:
    """Returns the `x` offset in pixels for selecting a region of interest."""

    return self._cam.get_offsetX()

  def _get_oy(self) -> int:
    """Returns the `y` offset in pixels for selecting a region of interest."""

    return self._cam.get_offsetY()

  def _get_extt(self) -> str:
    """Returns the current trigger mode value, and updates the last read
    trigger mode value if needed.

    The possible values for the trigger mode are `'Hardware'`, `'Free run'`,
    and `'Hdw after config'`.
    """

    r = self._cam.get_trigger_source()
    if r == 'XI_TRG_OFF' and self._trig == 'Hardware':
      self._trig = 'Free run'
    elif r != 'XI_TRG_OFF' and self._trig != 'Hardware':
      self._trig = 'Hardware'
    return self._trig

  def _get_framerate(self) -> float:
    """Returns the current framerate value of the camera."""

    return self._cam.get_framerate()

  def _get_framerate_mode(self) -> str:
    """Returns the current frame rate mode for the camera."""

    return FRAMERATE_MODES[self._model][self._cam.get_acq_timing_mode()]

  def _get_downsampling_mode(self) -> str:
    """Returns the current downsampling mode."""

    return DOWNSAMPLING_MODES[self._model][self._cam.get_downsampling()]

  def _set_data_format(self, fmt: str) -> None:
    """sets the requested data format."""

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    self._cam.set_imgdataformat(DATA_FORMATS_INV[self._model][fmt])

    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

  def _set_exp(self, exposure: float) -> None:
    """Sets the exposure time, in microseconds."""

    if self._get_aeag():
      self.log(logging.WARNING, "Setting the exposure won't work as long as "
                                "the AEAG is enabled !")
    self._cam.set_exposure(exposure)

  def _set_aeag(self, val: bool) -> None:
    """Enables or disables the Auto Exposure / Auto Gain setting."""

    if val:
      self._cam.enable_aeag()
    else:
      self._cam.disable_aeag()

  def _set_gain(self, gain: float) -> None:
    """Sets the gain, in dB."""

    if self._get_aeag():
      self.log(logging.WARNING, "Setting the gain won't work as long as the "
                                "AEAG is enabled !")
    self._cam.set_gain(gain)

  def _set_gamma_y(self, gamma: float) -> None:
    """Sets the Gamma Y value on the camera."""

    self._cam.set_gammaY(gamma)

  def _set_gamma_c(self, gamma: float) -> None:
    """Sets the Gamma C value on the camera."""

    self._cam.set_gammaC(gamma)
  
  def _set_sharpness(self, sharpness: float) -> None:
    """Sets the sharpness value on the camera."""

    self._cam.set_sharpness(sharpness)

  def _set_w(self, width: int) -> None:
    """Sets the width in pixels for selecting a region of interest."""

    # Lowering the X offset if it conflicts with the new image width
    if self.x_offset + width > self.settings['image_width'].highest:
      self.x_offset = self.settings['image_width'].highest - width

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    # Setting the requested width and reloading the X offset
    self._cam.set_width(width)

    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

    self.settings['x_offset'].reload(self._cam.get_offsetX_minimum(),
                                     self._cam.get_offsetX_maximum(),
                                     self._get_ox(),
                                     self._get_ox(),
                                     self._cam.get_offsetX_increment())

  def _set_h(self, height: int) -> None:
    """Sets the height in pixels for selecting a region of interest."""

    # Lowering the Y offset if it conflicts with the new image height
    if self.y_offset + height > self.settings['image_height'].highest:
      self.y_offset = self.settings['image_height'].highest - height

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    # Setting the requested height and reloading the Y offset
    self._cam.set_height(height)

    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

    self.settings['y_offset'].reload(self._cam.get_offsetY_minimum(),
                                     self._cam.get_offsetY_maximum(),
                                     self._get_oy(),
                                     self._get_oy(),
                                     self._cam.get_offsetY_increment())

  def _set_ox(self, x_offset: int) -> None:
    """Sets the `x` offset in pixels for selecting a region of interest."""

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    self._cam.set_offsetX(x_offset)

    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

  def _set_oy(self, y_offset: int) -> None:
    """Sets the `y` offset in pixels for selecting a region of interest."""

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    self._cam.set_offsetY(y_offset)

    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

  def _set_ext_trig(self, trig: str) -> None:
    """Sets the requested trigger mode value, and updates the last requested
    trigger mode value.

    The possible values for the trigger mode are `'Hardware'`, `'Free run'`,
    and `'Hdw after config'`.
    """

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    if trig == 'Hardware':
      self._cam.set_gpi_mode('XI_GPI_TRIGGER')
      self._cam.set_trigger_source('XI_TRG_EDGE_RISING')
    else:
      self._cam.set_gpi_mode('XI_GPI_OFF')
      self._cam.set_trigger_source('XI_TRG_OFF')
      
    self._trig = trig
    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

  def _set_framerate_mode(self, mode: str) -> None:
    """Sets the framerate mode for the camera to use."""

    self._cam.set_acq_timing_mode(FRAMERATE_MODES_INV[self._model][mode])

    self.settings['framerate'].reload(self._cam.get_framerate_minimum(),
                                      min(self._cam.get_framerate_maximum(),
                                          500),
                                      self._cam.get_framerate(),
                                      self._cam.get_framerate(),
                                      self._cam.get_framerate_increment())

  def _set_framerate(self, framerate) -> None:
    """Sets the target framerate value of the camera."""

    if self._get_framerate_mode() == 'Free run':
      self.log(logging.WARNING, "Setting the framerate won't work as long as "
                                "the camera is in free run mode !")
    self._cam.set_framerate(framerate)

  def _set_downsampling_mode(self, mode: str) -> None:
    """Sets the downsampling mode on the camera."""

    if self._started:
      self.log(logging.DEBUG, "Stopping the image acquisition")
      self._cam.stop_acquisition()

    self._cam.set_downsampling(DOWNSAMPLING_MODES_INV[self._model][mode])

    if self._started:
      self.log(logging.DEBUG, "Starting the image acquisition")
      self._cam.start_acquisition()

    self.settings['image_width'].reload(self._cam.get_width_minimum(),
                                        self._cam.get_width_maximum(),
                                        self._get_w(),
                                        self._get_w(),
                                        self._cam.get_width_increment())

    self.settings['image_height'].reload(self._cam.get_height_minimum(),
                                         self._cam.get_height_maximum(),
                                         self._get_h(),
                                         self._get_h(),
                                         self._cam.get_height_increment())

    self.settings['x_offset'].reload(self._cam.get_offsetX_minimum(),
                                     self._cam.get_offsetX_maximum(),
                                     self._get_ox(),
                                     self._get_ox(),
                                     self._cam.get_offsetX_increment())

    self.settings['y_offset'].reload(self._cam.get_offsetY_minimum(),
                                     self._cam.get_offsetY_maximum(),
                                     self._get_oy(),
                                     self._get_oy(),
                                     self._cam.get_offsetY_increment())



# coding: utf-8

from time import time
from typing import List, Optional
import logging

from ..meta_inout import InOut
from ...tool.ft232h import FT232HServer as FT232H, USBArgsType

# Register and other configuration values:
Ads1115_pointer_conversion = 0x00
Ads1115_pointer_config = 0x01
Ads1115_pointer_lo_thresh = 0x02
Ads1115_pointer_hi_thresh = 0x03

Ads1115_config_mux = {'A0 - A1': 0x0000,
                      'A0 - A3': 0x1000,
                      'A1 - A3': 0x2000,
                      'A2 - A3': 0x3000,
                      'A0': 0x4000,
                      'A1': 0x5000,
                      'A2': 0x6000,
                      'A3': 0x7000}

Ads1115_config_gain = {0.256: 0x0a00,
                       0.512: 0x0800,
                       1.024: 0x0600,
                       2.048: 0x0400,
                       4.096: 0x0200,
                       6.144: 0x0000}

Ads1115_config_dr = {8: 0x0000,
                     16: 0x0020,
                     32: 0x0040,
                     64: 0x0060,
                     128: 0x0080,
                     250: 0x00A0,
                     475: 0x00C0,
                     860: 0x00E0}


class ADS1115FT232H(InOut):
  """A class for controlling Adafruit's ADS1115 16-bits ADC through an FT232H.

  It is similar to the :class:`~crappy.inout.ADS1115` class, except this class
  is specific for use with an :class:`~crappy.tool.ft232h.FT232H` USB to I2C
  converter.

  The ADS1115 InOut is meant for reading conversion values from a 16-bits
  ADS1115 ADC, using the I2C protocol. The output is in Volts by default, but a
  ``gain`` and an ``offset`` can be specified.

  Various settings can be adjusted, like the sample rate, the input mode or the
  voltage range.
  
  .. versionadded:: 2.0.0
  """

  ft232h = True

  def __init__(self,
               device_address: int = 0x48,
               sample_rate: int = 128,
               v_range: float = 2.048,
               multiplexer: str = 'A1',
               dry_pin: Optional[str] = None,
               gain: float = 1,
               offset: float = 0,
               _ft232h_args: USBArgsType = tuple()) -> None:
    """Checks the validity of the arguments.

    Args:
      device_address: The I2C address of the ADS1115. The default address is
        `0x48`, but it is possible to change this setting using the `ADDR` pin.
      sample_rate: The sample rate for data conversion (in SPS). The available
        sample rates are:
        ::

          8, 16, 32, 64, 128, 250, 475, 860

      v_range: The value (in Volts) of the measured signal corresponding to the
        `0x7FFF` output in bits, i.e. that saturates the sensor. A signal of
        ``-v_range`` Volts gives a `0x8000` output in bits. Available
        ``v_range`` values are:
        ::

          0.256, 0.512, 1.024, 2.048, 4.096, 6.144

      multiplexer: Choice of the inputs to consider. Single-input modes
        actually measure `Ax - GND`. The available ``multiplexer`` values are:
        ::

          'A0', 'A1', 'A2', 'A3',
          'A0 - A1',
          'A0 - A3',
          'A1 - A3',
          'A2 - A3'

      dry_pin: Optionally, reads the end of conversion signal from a GPIO
        rather than from an I2C message. Speeds up the reading and decreases
        the traffic on the I2C bus, but requires one extra wire. Give the name
        of the GPIO in the format `Dx` or `Cx`.
      gain: Allows to tune the output value according to the formula :
        :math:`output = gain * tension + offset`.
      offset: Allows to tune the output value according to the formula :
        :math:`output = gain * tension + offset`.
      _ft232h_args: This argument is meant for internal use only and should not
        be provided by the user. It contains the information necessary for
        setting up the FT232H.

    Warning:
      AINx voltages should not be higher than `VDD+0.3V` nor lower than
      `GND-0.3V`. Setting high ``v_range`` values does not allow measuring
      voltages higher than `VDD` !!
    """

    self._bus = None

    super().__init__()

    (block_index, block_lock, command_file, answer_file, shared_lock,
     current_block) = _ft232h_args

    self._bus = FT232H(mode='I2C',
                       block_index=block_index,
                       current_block=current_block,
                       command_file=command_file,
                       answer_file=answer_file,
                       block_lock=block_lock,
                       shared_lock=shared_lock)

    self._device_address = device_address

    if v_range not in Ads1115_config_gain:
      raise ValueError("v_range should be in {}".format(list(
        Ads1115_config_gain.keys())))
    else:
      self._v_range = v_range

    if sample_rate not in Ads1115_config_dr:
      raise ValueError("sample_rate should be in {}".format(list(
        Ads1115_config_dr.keys())))
    else:
      self._sample_rate = sample_rate

    if multiplexer not in Ads1115_config_mux:
      raise ValueError("multiplexer should be in {}".format(list(
        Ads1115_config_mux.keys())))
    else:
      self._multiplexer = multiplexer

    if dry_pin is not None and not isinstance(dry_pin, str):
      raise TypeError('int_pin should be a string !')
    self._dry_pin = dry_pin

    self._gain = gain
    self._offset = offset

  def open(self) -> None:
    """Initializes the I2C communication and the device."""

    if not self._is_connected():
      raise IOError("The ADS1115 is not connected")

    # Setting the configuration register according to the user values
    init_value = Ads1115_config_mux[self._multiplexer]
    init_value |= Ads1115_config_gain[self._v_range]
    init_value |= 0x0100  # Single shot operating mode
    init_value |= Ads1115_config_dr[self._sample_rate]
    self.log(logging.INFO, "Initializing the ADS1115")
    if self._dry_pin is None:
      # Setting the two last bits to 11 to disable the DRY pin
      init_value |= 0x0003

    self._set_register(Ads1115_pointer_config, init_value)

    # Setting the threshold registers to activate the DRY pin output
    if self._dry_pin is not None:
      self._set_register(Ads1115_pointer_lo_thresh, 0x0000)
      self._set_register(Ads1115_pointer_hi_thresh, 0xFFFF)

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in Volts, unless a gain and offset are applied.

    Returns:
      A :obj:`list` containing the timestamp and the voltage value.
    """

    # Reading the config register, and setting it so that the ADS1115 starts
    # a conversion
    ms_byte, ls_byte = self._bus.read_i2c_block_data(self._device_address,
                                                     Ads1115_pointer_config,
                                                     2)
    self._set_register(Ads1115_pointer_config,
                       ((ms_byte | 0x80) << 8) | ls_byte)

    # Waiting for the end of the conversion
    t0 = time()
    while not self._data_available():
      if time() - t0 > 0.5:
        raise TimeoutError('Waited too long for data to be ready !')

    # Reading the output of the conversion
    out = [time()]
    ms_byte, ls_byte = self._bus.read_i2c_block_data(
      self._device_address,
      Ads1115_pointer_conversion,
      2)
    self.log(logging.DEBUG, f"Read {ms_byte, ls_byte} from the device "
                            f"address {self._device_address} at "
                            f"register {Ads1115_pointer_conversion}")

    # Converting the output value into Volts
    value_raw = (ms_byte << 8) | ls_byte
    if ms_byte >> 7:
      value_raw -= 2 ** 16
    value = value_raw * self._v_range / 2 ** 15

    out.append(self._offset + self._gain * value)
    return out

  def close(self) -> None:
    """Closes the I2C bus."""

    if self._bus is not None:
      self.log(logging.INFO, "Closing the ADS1115")
      self._bus.close()

  def _set_register(self, register_address: int, value: int) -> None:
    """Thin wrapper for writing data to the registers."""

    self.log(logging.DEBUG, f"Writing {[(value >> 8) & 0xFF, value & 0xFF]} "
                            f"to the address {self._device_address} in "
                            f"register {register_address}")
    self._bus.write_i2c_block_data(self._device_address, register_address,
                                   [(value >> 8) & 0xFF, value & 0xFF])

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns:
      :obj:`True` if reading was successful, else :obj:`False`.
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._dry_pin is None:
      return bool(self._bus.read_i2c_block_data(self._device_address,
                                                Ads1115_pointer_config,
                                                1)[0] & 0x80)
    # EOC signal from a GPIO
    else:
      return not bool(self._bus.get_gpio(self._dry_pin))



# coding: utf-8

from typing import Optional, Tuple
import numpy as np
import logging
from  warnings import warn

from . import BaslerIronmanCameraLink
from ..._global import OptionalModule

try:
  from . import clModule as Cl
except (ModuleNotFoundError, ImportError):
  Cl = OptionalModule("clModule")

format_to_num = {'8 bits': 0,
                 '10 bits': 1,
                 '12 bits': 2}
num_to_format = {val: key for key, val in format_to_num.items()}


class JaiGO5000CPMCL8Bits(BaslerIronmanCameraLink):
  """This class can drive a JAI GO-5000C-PMCL camera in 8 bits mode, through a
  Basler microEnable 5 Ironman AD8 PoCL acquisition board.

  It is a child of the
  :class:`~crappy.camera.cameralink.BaslerIronmanCameraLink` Camera. The only
  difference is that it can only run in 8 bits mode, and gives access to more
  camera settings in Crappy. It is subclassed by the
  :class:`~crappy.camera.cameralink.JaiGO5000CPMCL` Camera, that allows driving
  the same hardware but in 10 and 12 bits mode.

  Warning:
    This Camera relies on a custom-written C library that hasn't been tested in
    a long time. It might not be functional anymore. This Camera also requires
    proprietary drivers to be installed.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Jai8 to JaiGO5000CPMCL8Bits
  .. versionremoved:: 2.1.0
  """

  def __init__(self) -> None:
    """Adds various settings to the Camera."""

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be "
         f"deprecated and removed from Crappy. Please contact the maintainers "
         f"if you still use this Camera.", FutureWarning)

    super().__init__()

    self.add_scale_setting('width', 1, 2560, self._get_w, self._set_w, 2560)
    self.add_scale_setting('height', 1, 2048, self._get_h, self._set_h, 2048)
    self.add_scale_setting('exposure', 10, 800000, self._get_exp,
                           self._set_exp)

  def open(self,
           config_file: Optional[str] = None,
           camera_type: str = 'FullAreaGray8',
           **kwargs) -> None:
    """Opens the connection to the camera using the parent class' method, and
    sets the image format to 8 bits.

    Args:
      config_file: Path to the configuration file for the camera, as a
        :obj:`str`. Allows setting various parameters at once, and to store
        them in a persistent way.
      camera_type: The type of camera to drive, as a :obj:`str`.
      **kwargs: All the settings to set on the camera.

    .. versionadded:: 1.5.10
       explicitly listing the *config_file* and *camera_type* arguments
    """

    super().open(config_file=config_file,
                 camera_type=camera_type,
                 **kwargs)

    self.log(logging.DEBUG, "Writing b'TAGM=5\\r\\n' to the camera")
    self._cap.serialWrite('TAGM=5\r\n')  # (default)
    self._set_format('8 bits')  # Set camera to 8 bits
    self._cap.set(Cl.FG_CAMERA_LINK_CAMTYP, 208)  # Set the input to 8 bits
    self._cap.set(Cl.FG_SENSORREADOUT, 0)  # Sets the correct framegrabber mode

  def _set_w(self, val: int) -> None:
    self.stopAcq()
    super()._set_w(val)
    self.log(logging.DEBUG, f"Writing b'WTC={val}\\r\\n' to the camera")
    self._cap.serialWrite('WTC={}\r\n'.format(val))
    self.startAcq()

  def _set_h(self, val: int) -> None:
    self.stopAcq()
    super()._set_h(val)
    self.log(logging.DEBUG, f"Writing b'HTL={val}\\r\\n' to the camera")
    self._cap.serialWrite('HTL={}\r\n'.format(val))
    self.startAcq()

  def _get_format(self) -> str:
    self.log(logging.DEBUG, "Writing b'BA?\\r\\n' to the camera")
    r = self._cap.serialWrite('BA?\r\n')
    return num_to_format[int(r[3])]

  def _set_format(self, val: str) -> None:
    self.log(logging.DEBUG, f"Writing b'BA={format_to_num[val]}\\r\\n' "
                            f"to the camera")
    self._cap.serialWrite('BA={}\r\n'.format(format_to_num[val]))

  def _set_exp(self, val: int) -> None:
    self.log(logging.DEBUG, f"Writing b'RE={val}\\r\\n' to the camera")
    self._cap.serialWrite('PE={}\r\n'.format(val))

  def _get_exp(self) -> int:
    self.log(logging.DEBUG, "Writing b'PE?\\r\\n' to the camera")
    return int(self._cap.serialWrite('PE?\r\n').strip()[3:])


class JaiGO5000CPMCL(JaiGO5000CPMCL8Bits):
  """This class can drive a JAI GO-5000C-PMCL camera in 10 or 12 bits mode,
  through a Basler microEnable 5 Ironman AD8 PoCL acquisition board.

  It is a child of the :class:`~crappy.camera.cameralink.JaiGO5000CPMCL8Bits`
  Camera. The only difference with its parent class is that it sets the 10 or
  12 bits mode on the camera, and modifies the acquired image before returning
  it.
  
  Warning:
    This Camera relies on a custom-written C library that hasn't been tested in
    a long time. It might not be functional anymore. This Camera also requires
    proprietary drivers to be installed.

  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Jai to JaiGO5000CPMCL
  .. versionremoved:: 2.1.0
  """

  def __init__(self) -> None:
    """Adds the data_format settings to the Camera."""

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be "
         f"deprecated and removed from Crappy. Please contact the maintainers "
         f"if you still use this Camera.", FutureWarning)

    super().__init__()
    self.add_choice_setting('data_format', ('10 bits', '12 bits'),
                            self._get_format, self._set_format, '12 bits')

  def open(self,
           camera_type: str = 'MediumAreaGray16',
           **kwargs) -> None:
    """Opens the connection to the camera using the parent class' method, and
    sets the image format to 12 bits.

    Args:
      camera_type: The type of camera to drive, as a :obj:`str`.
      **kwargs: All the settings to set on the camera.

    .. versionadded:: 1.5.10 explicitly listing the *camera_type* argument
    """

    super().open(camera_type=camera_type, **kwargs)
    # dual tap (default does not allow 12 bits)
    self.log(logging.DEBUG, "Writing b'TAGM=1\\r\\n' to the camera")
    self._cap.serialWrite('TAGM=1\r\n')
    self._set_format('12 bits')  # 12 bits
    self._cap.set(Cl.FG_CAMERA_LINK_CAMTYP, 212)  # Set the input to 12 bits
    self._cap.set(Cl.FG_SENSORREADOUT, 7)  # Sets the correct framegrabber mode

  def get_image(self) -> Tuple[float, np.ndarray]:
    """Grabs a frame using the parent class' method, and returns if after
    shifting bits."""

    t, frame = super().get_image()
    return t, frame >> 4



# coding: utf-8

from time import time
from numpy import sin, pi
from typing import Union, Dict
import logging

from .meta_path import Path, ConditionType


class Sine(Path):
  """This Path generates a sine wave varying with time until the stop condition
  is met.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               condition: Union[str, ConditionType],
               freq: float,
               amplitude: float,
               offset: float = 0,
               phase: float = 0) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      condition: The condition for switching to the next Path. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      freq: The frequency of the sine in `Hz`.
      amplitude: The amplitude of the sine wave (peak to peak).
      offset: The offset of the sine (average value).
      phase: The phase of the sine (in radians).
    
    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    super().__init__()

    # Setting the attributes
    self._condition = self.parse_condition(condition)
    self._amplitude = amplitude / 2
    self._offset = offset
    self._phase = phase
    self._k = 2 * pi * freq

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met."""

    # Checking if the stop condition is met
    if self._condition(data):
      self.log(logging.DEBUG, "Stop condition met")
      raise StopIteration

    # Returning the current signal value
    return sin((time() - self.t0) * self._k - self._phase) * \
        self._amplitude + self._offset



# coding: utf-8

import numpy as np
from typing import Optional, Union, Iterable
from time import time
import logging

from .meta_block import Block


class MeanBlock(Block):
  """This Block can compute the average values of given labels over a given 
  delay.

  It can take any number of inputs, provided that they share a common time
  label. If the same label (except time) is received from several Blocks, it
  may lead to unexpected results.
  
  The output of this Block is very similar to that of the 
  :class:`~crappy.modifier.Mean` and :class:`~crappy.modifier.MovingAvg` 
  Modifiers, but not exactly similar. While these Modifiers calculate the 
  average of a label over a fixed number of data points, the MeanBlock 
  calculates the average of the values received over a given delay. This 
  behavior could, however, also be achieved using a 
  :class:`~crappy.modifier.Modifier`.

  Warning:
    If the delay for averaging is too short compared with the looping frequency
    of the upstream Blocks, this Block may not always return the same number of
    labels ! This can cause errors in downstream Blocks expecting a fixed
    number of labels.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Mean_block to MeanBlock
  """

  def __init__(self,
               delay: float,
               time_label: str = 't(s)',
               out_labels: Optional[Union[str, Iterable[str]]] = None,
               display_freq: bool = False,
               freq: Optional[float] = 50,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      delay: The averaged data will be sent each ``delay`` seconds.
      time_label: The label containing the time information. It must be common
        to all the incoming Links.
      out_labels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        all the labels to average, as :obj:`str`. If not given, all the
        received labels are averaged and returned. The time label should not
        be included, as it is already given in ``time_label``. If there is only
        one label to average, it can be directly given as a :obj:`str`, i.e.
        not in an iterable.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded:: 1.5.10 *verbose* argument
    .. versionchanged:: 1.5.10 renamed *t_label* argument to *time_label*
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__()
    self.display_freq = display_freq
    self.freq = freq
    self.debug = debug

    self._delay = delay
    self._time_label = time_label

    # Forcing the out_labels into a list
    if out_labels is not None and isinstance(out_labels, str):
      self._out_labels = [out_labels]
    elif out_labels is not None:
      self._out_labels = list(out_labels)
    else:
      self._out_labels = None

    self._last_sent_t = time()

  def begin(self) -> None:
    """Initializes the time counter.
    
    .. versionadded:: 2.0.0
    """

    self._last_sent_t = self.t0

  def loop(self) -> None:
    """Receives all available data from the upstream Blocks, averages it and
    sends it if the time delay is reached."""

    # Receiving data from each incoming link
    data = self.recv_all_data(delay=self._delay, poll_delay=self._delay / 10)
    to_send = dict()

    # Removing the time label from the received data
    if self._time_label in data:
      data.pop(self._time_label)

    # Building the output dict with the averaged values
    for label, values in data.items():
      if self._out_labels is None or label in self._out_labels:
        try:
          to_send[label] = np.mean(values)
        except (ValueError, TypeError):
          self.log(logging.WARNING, f"Cannot perform averaging on label "
                                    f"{label} with values: {values}")
          to_send[label] = values[-1]

    # Sending the output dict
    if to_send:
      to_send[self._time_label] = (time() + self._last_sent_t) / 2 - self.t0
      self._last_sent_t = time()
      self.send(to_send)



# coding: utf-8

from threading import Thread
from math import log2, ceil
import numpy as np
from typing import Optional, Iterable
from time import time, sleep
import logging
import logging.handlers

from .camera_process import CameraProcess
from ..._global import OptionalModule
from ...tool.camera_config import Overlay

plt = OptionalModule('matplotlib.pyplot', lazy_import=True)

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class Displayer(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can display 
  images acquired by a :class:`~crappy.blocks.Camera` Block in a dedicated 
  window.
  
  It is meant to serve as a control or validation feature, its resolution is
  thus limited to `640x480` and it should not be used at high framerates. On
  top of the displayed image, it can also draw 
  :class:`~crappy.tool.camera_config.config_tools.Overlay` objects sent by
  other :class:`~crappy.blocks.camera_processes.CameraProcess`. This way, the
  user can for example visualize the spots being tracked by the
  :class:`~crappy.blocks.VideoExtenso` Block in real time.

  The images can be displayed using two different backends : either using
  :mod:`cv2` (OpenCV), or using :mod:`matplotlib`. OpenCV is by far the fastest
  and most convenient.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               title: str,
               framerate: float,
               backend: Optional[str] = None) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      title: The name of the Displayer window, that will be displayed on the
        window border.
      framerate: The target framerate for the display. The actual achieved
        framerate might be lower, but never greater than this value.
      backend: The module to use for displaying the images. Can be either
        ``'cv2'`` or ``'mpl'``, to use respectively :mod:`cv2` or
        :mod:`matplotlib`.
    """

    # The thread must be initialized later for compatibility with Windows
    self._overlay_thread: Optional[Thread] = None
    self._overlay: Iterable[Overlay] = list()
    self._stop_thread = False

    super().__init__()

    self._title = title
    self._framerate = framerate

    # Selecting the backend if no backend was specified
    if backend is None:
      if not isinstance(cv2, OptionalModule):
        self._backend = 'cv2'
      else:
        try:
          _ = plt.Figure
          self._backend = 'mpl'
        except RuntimeError:
          raise ModuleNotFoundError("Neither opencv-python nor matplotlib "
                                    "could be imported, no backend found for "
                                    "displaying the images")

    elif backend in ('cv2', 'mpl'):
      self._backend = backend
    else:
      raise ValueError("The backend argument should be either 'cv2' or "
                       "'mpl' !")

    # Setting other instance attributes
    self._ax = None
    self._fig = None
    self._last_upd = time()

  def __del__(self) -> None:
    """On exit, ensuring that the :obj:`~threading.Thread` in charge of
    grabbing the :class:`~crappy.tool.camera_config.config_tools.Overlay` to
    display has stopped, otherwise stopping it."""

    if self._overlay_thread is not None and self._overlay_thread.is_alive():
      self._stop_thread = True
      try:
        self._overlay_thread.join(0.05)
      except RuntimeError:
        pass

  def init(self) -> None:
    """Starts the :obj:`~threading.Thread` for grabbing the
    :class:`~crappy.tool.camera_config.config_tools.Overlay` to display, and
    initializes the Displayer window."""

    # Instantiating and starting the Thread for grabbing the Overlays
    self.log(logging.INFO, "Instantiating the thread for getting the Overlays"
                           " to display")
    self._overlay_thread = Thread(target=self._thread_target)
    self.log(logging.INFO, "Starting the thread for getting the Overlays to "
                           "display")
    self._overlay_thread.start()

    # Preparing the Displayer window
    self.log(logging.INFO, f"Opening the displayer window with the backend "
                           f"{self._backend}")
    if self._backend == 'cv2':
      self._prepare_cv2()
    elif self._backend == 'mpl':
      self._prepare_mpl()

  def _get_data(self) -> bool:
    """Method similar to the one of the parent class, except it also ensures 
    that the achieved framerate stays within the limit specified by the user.
    
    Returns:
      :obj:`True` in case a frame was acquired and needs to be handled, or
      :obj:`False` if no frame was grabbed and nothing should be done.
    """

    # Acquiring the Lock to avoid conflicts with other CameraProcesses
    with self._lock:

      # In case there's no frame grabbed yet
      if 'ImageUniqueID' not in self._data_dict:
        return False

      # In case the frame in buffer was already handled during a previous loop,
      # or it's too early to grab a new frame because of the target framerate
      if self._data_dict['ImageUniqueID'] == self.metadata['ImageUniqueID'] \
          or time() - self._last_upd < 1 / self._framerate:
        return False

      # Copying the metadata
      self.metadata = self._data_dict.copy()
      self._last_upd = time()

      self.log(logging.DEBUG, f"Got new image to process with id "
                              f"{self.metadata['ImageUniqueID']}")

      # Copying the frame
      np.copyto(self.img,
                np.frombuffer(self._img_array.get_obj(),
                              dtype=self._dtype).reshape(self._shape))

    return True

  def loop(self) -> None:
    """This method grabs the latest frame, casts it to 8 bits if necessary,
    and updates the Displayer window to draw it.
    
    It also draws the latest received 
    :class:`~crappy.tool.camera_config.config_tools.Overlay` on top of the
    displayed frame.
    """

    # Casting the image to uint8 if it's not already in this format
    if self.img.dtype != np.uint8:
      self.log(logging.DEBUG, f"Casting displayed image from "
                              f"{self.img.dtype} to uint8")
      if np.max(self.img) > 255:
        factor = max(ceil(log2(np.max(self.img) + 1) - 8), 0)
        img = (self.img / 2 ** factor).astype(np.uint8)
      else:
        img = self.img.astype(np.uint8)
    else:
      img = self.img.copy()

    # Drawing the latest known overlay
    for overlay in self._overlay:
      if overlay is not None:
        self.log(logging.DEBUG, f"Drawing {overlay} on top of the image to "
                                "display")
        overlay.draw(img)

    # Calling the right update method
    if self._backend == 'cv2':
      self._update_cv2(img)
    elif self._backend == 'mpl':
      self._update_mpl(img)

  def finish(self) -> None:
    """Closes the Displayer window and stops the :obj:`~threading.Thread`
    grabbing the :class:`~crappy.tool.camera_config.config_tools.Overlay`"""

    # Closing the Displayer window
    self.log(logging.INFO, "Closing the displayer window")
    if self._backend == 'cv2':
      self._finish_cv2()
    elif self._backend == 'mpl':
      self._finish_mpl()

    # Stooping the Thread grabbing the Overlay to draw
    if self._overlay_thread is not None and self._overlay_thread.is_alive():
      self._stop_thread = True
      try:
        self._overlay_thread.join(0.05)
      except RuntimeError:
        self.log(logging.WARNING, "Thread for receiving the Overlay did not "
                                  "stop as expected")

  def _thread_target(self) -> None:
    """This method is the target to the :obj:`~threading.Thread` in charge of
    grabbing the :class:`~crappy.tool.camera_config.config_tools.Overlay` to
    draw on top of the displayed image.
    
    It repeatedly polls the :obj:`~multiprocessing.Connection` through which
    the Overlays are received, and stores the last received Overlays.
    """

    # Looping until the entire CameraProcess is told to stop, or the 
    # _stop_thread flag is raised
    while not self._stop_event.is_set() and not self._stop_thread:

      # Receiving the latest Overlay to draw
      overlay = None
      while self._to_draw_conn.poll():
        overlay = self._to_draw_conn.recv()

      # Saving the received Overlay
      if overlay is not None:
        self.log(logging.DEBUG, f"Received overlay to display: {overlay}")
        self._overlay = overlay

      # To avoid spamming the CPU in vain
      else:
        sleep(0.001)

    self.log(logging.INFO, "Thread for receiving the Overlays ended")

  def _prepare_cv2(self) -> None:
    """Instantiates the display window of :mod:`cv2`."""

    try:
      flags = cv2.WINDOW_NORMAL | cv2.WINDOW_KEEPRATIO
    except AttributeError:
      flags = cv2.WINDOW_NORMAL
    cv2.namedWindow(self._title, flags)

  def _prepare_mpl(self) -> None:
    """Creates a :mod:`matplotlib` Figure."""

    plt.ion()
    self._fig, self._ax = plt.subplots()

  def _update_cv2(self, img: np.ndarray) -> None:
    """Reshapes the image to a maximum shape of 640x480 and displays it in 
    :mod:`cv2`."""

    if img.shape[0] > 480 or img.shape[1] > 640:
      factor = min(480 / img.shape[0], 640 / img.shape[1])
      self.log(logging.DEBUG,
               f"Reshaping displayed image from {img.shape} to "
               f"{int(img.shape[1] * factor), int(img.shape[0] * factor)}")
      img = cv2.resize(img, (int(img.shape[1] * factor),
                             int(img.shape[0] * factor)))

    self.log(logging.DEBUG, "Displaying the image")
    cv2.imshow(self._title, img)
    cv2.waitKey(1)

  def _update_mpl(self, img: np.ndarray) -> None:
    """Reshapes the image to a dimension inferior or equal to 640x480 and
    displays it in :mod:`matplotlib`."""

    if img.shape[0] > 480 or img.shape[1] > 640:
      factor = max(ceil(img.shape[0] / 480), ceil(img.shape[1] / 640))
      self.log(logging.DEBUG,
               f"Reshaping the displayed image from {img.shape} to "
               f"{(img.shape[0] / factor, img.shape[1] / factor)}")
      img = img[::factor, ::factor]

    self._ax.clear()
    self.log(logging.DEBUG, "Displaying the image")
    self._ax.imshow(img, cmap='gray')
    plt.pause(0.001)
    plt.show()

  def _finish_cv2(self) -> None:
    """Destroys the opened :mod:`cv2` window."""

    if self._title is not None:
      cv2.destroyWindow(self._title)

  def _finish_mpl(self) -> None:
    """Destroys the opened :mod:`matplotlib` window."""

    if self._fig is not None:
      plt.close(self._fig)



# coding: utf-8

from time import time
from typing import List, Optional, Dict, Any, Union, Tuple, Iterable
from itertools import chain
from dataclasses import dataclass, field
from multiprocessing import current_process
import logging

from .meta_inout import InOut
from .._global import OptionalModule

try:
  from labjack import ljm
except (ModuleNotFoundError, ImportError):
  ljm = OptionalModule("ljm", "Please install Labjack LJM and the ljm "
                              "Python module")

# Map of thermocouple types vs indexes for the Labjack T7
thcp_map = {'E': 20, 'J': 21, 'K': 22, 'R': 23, 'T': 24, 'S': 25, 'C': 30}


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a Labjack
  channel can have.

  Not all the attributes are used by every channel, but they all have a use for
  at least one type of channel.
  """

  name: Union[str, int]

  direction: bool = True
  dtype: Optional[int] = None
  address: int = 1
  gain: float = 1
  offset: float = 0
  make_zero: bool = False
  range: float = 10
  limits: Optional[Tuple[float, float]] = None
  resolution: int = 1
  thermocouple: Optional[str] = None
  write_at_open: List[Tuple[str, float]] = field(default_factory=list)

  def update(self, dic_in: Dict[str, Any]) -> None:
    """Updates the channel keys based on the user input."""

    for key, val in dic_in.items():
      if hasattr(self, key):
        setattr(self, key, val)

      # Handling the case when the user enters a wrong key
      else:
        logger = logging.getLogger(
          f"{current_process().name}.LabjackT7.Channel_{self.name}")
        logger.log(logging.WARNING, f"Unknown channel key : {key}, ignoring")


class LabjackT7(InOut):
  """This InOut allows controlling a Labjack T7 device. It can use any channel
  as input/output.

  The Labjack T7 is a very complete DAQ board. It features several ADC, several
  DAC, as well as multiple GPIOs. It can also read thermocouples, and run LUA
  code on an integrated microcontroller. These features can all be controlled
  from Crappy.

  This class is not capable of streaming. For higher frequency, refer to the
  :class:`~crappy.inout.T7Streamer` class.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Labjack_t7 to LabjackT7
  """

  def __init__(self,
               channels: Iterable[Dict[str, Any]],
               device: str = 'ANY',
               connection: str = 'ANY',
               identifier: str = 'ANY',
               write_at_open: Optional[Iterable[tuple]] = None,
               no_led: bool = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      channels: An iterable (like a :obj:`list` or a :obj:`tuple`) of the
        channels to interface with on the Labjack. Each object in this iterable
        should be a :obj:`dict` representing a single channel, and whose keys
        provide information on the channel to use. Refer to the note below for
        more information on the possible keys.
      device: The type of Labjack to open. Possible values include :
        ::

          'ANY', 'T7', 'T4', 'DIGIT'

        Only tested with `'T7'` in Crappy.
      connection: The type of connection used for interfacing with the Labjack.
        Possible values include :
        ::

          'ANY', 'TCP', 'USB', 'ETHERNET', 'WIFI'

      identifier: Any extra information allowing to further identify the
        Labjack to open, like a serial number, an IP address, or a device name.
      write_at_open: If specific names or registers have to be written when
        opening the channel, they can be given here as an iterable (like a
        :obj:`list` or a :obj:`tuple`) of :obj:`tuple`. They will be written in
        the same order as in the given iterable. Refer to the note below for
        the accepted formats.
      no_led: If :obj:`True`, turns off the LED on the Labjack. This led can
        generate noise on the channels `AIN0` and `AIN1`.

    Note:
      - ``channels`` keys:

        - name: The name of the channel to interface with, as written on the
          Labjack's case. Ex: `'AIN0'`. The available settings as well as the
          direction of the channel depend on the given name.

          The name can be:
            - `AINx`: Analog input, linked to an ADC. A gain and offset can be
              provided, and the range and resolution can be adjusted. It is
              always an input. These channels can also be used with
              thermocouples (see below).
            - `(T)DACx`: Analog output, linked to a DAC. A gain and an offset
              can be specified. It is always an output.
            - `(E/F/C/M IOx)`: Digital inputs/outputs. A gain and an offset
              can be specified. It can be either an input or an output, the
              default is output.

        - gain: If the channel is an input, the measured value will be
          modified directly by the Labjack as follows :
          :math:`returned\_value = gain * measured\_value + offset`. If the
          channel is an output, the command value will be modified in Crappy as
          follows before being sent to the Labjack :
          :math:`sent\_value = gain * command + offset`.
        - offset: If the channel is an input, the measured value will be
          modified directly by the Labjack as follows :
          :math:`returned\_value = gain * measured\_value + offset`. If the
          channel is an output, the command value will be modified in Crappy as
          follows before being sent to the Labjack :
          :math:`sent\_value = gain * command + offset`.
        - make_zero: If :obj:`True`, data will be acquired on this channel
          before the test starts, and a compensation value will be deduced
          so that the offset of this channel is `0`. The compensation is
          performed directly by the Labjack. This setting only has effect for
          `AIN` channels defined as inputs. **It will only take effect if the**
          ``make_zero_delay`` **argument of the**
          :class:`~crappy.blocks.IOBlock` **controlling the Labjack is set** !

        - direction: If :obj:`True`, the channel is considered as an output,
          else as an input. Only has effect for `IO` channels, the default is
          output.

        - resolution: The resolution of the acquisition as an integer, refer to
          Labjack documentation for more details. The higher this value the
          better the resolution, but the lower the speed. The possible range is
          either `1` to `8` or to `12` depending on the model. The default is
          `1`. Only has effect for `AIN` channels.

        - range: The range of the acquisition in Volts. A range of `x` means
          that values can be read  between `-x` and `x` Volts. The possible
          values are :
          ::

            0.01, 0.1, 1, 10

          Only has effect for `AIN` channels.

        - limits: A :obj:`tuple` containing the minimum and maximum allowed
          command values to set for this channel. After applying the gain and
          offset to the command, it is then clamped between these two values
          before being sent to the Labjack. Only has effect for output
          channels.

        - thermocouple: The type of thermocouple to read data from. Possible
          values are:
          ::

            'E', 'J', 'K', 'R', 'T', 'S', 'C'

          If specified, it will use the EF to read a temperature directly from
          the thermocouples. Only has effect for `AIN` channels.

        - write_at_open: A :obj:`list` containing commands for writing specific
          names or registers when opening the channel. The commands should be
          given as :obj:`tuple` either in the format
          `(name (str), value (int/float))` or
          `(register (int), type (int), value (float/int))`.

    Warning:
      Do not consider the ``limits`` key as a safety feature. It *should* not
      go beyond/below the given values, but this is not meant to replace
      hardware safety !
    """

    self._handle = None

    super().__init__()

    # Identifiers for the device to open
    self._device = device
    self._connection = connection
    self._identifier = identifier

    # List of commands to send when opening the Labjack
    self._write_at_open = [] if write_at_open is None else list(write_at_open)
    if no_led:
      self._write_at_open.append(('POWER_LED', 0))

    self._channels_in = list()
    self._channels_out = list()

    # Parsing the setting dict given for each channel
    for channel in channels:

      # Checking that the name was given as it's the most important attribute
      if 'name' not in channel:
        raise AttributeError("The given channels must contain the 'name' "
                             "key !")
      name = channel['name']

      # Modbus registers
      if isinstance(name, int):
        chan = _Channel(name=name, dtype=ljm.constants.FLOAT32)
        chan.update(channel)

        # Can be either input or output, the user has to specify
        if chan.direction:
          self._channels_out.append(chan)
        else:
          self._channels_in.append(chan)

      # Analog inputs
      elif isinstance(name, str) and name.startswith('AIN'):
        chan = _Channel(name=name)
        chan.update(channel)

        # Setting the range and the resolution
        chan.write_at_open.extend([
          (*self._parse(f'{name}_RANGE'), chan.range),
          (*self._parse(f'{name}_RESOLUTION_INDEX'), chan.resolution)])

        # Specific commands for thermocouples
        if chan.thermocouple is not None:
          chan.write_at_open.extend([
            (*self._parse(f'{name}_EF_INDEX'), thcp_map[chan.thermocouple]),
            (*self._parse(f'{name}_EF_CONFIG_A'), 1),
            (*self._parse(f'{name}_EF_CONFIG_B'), 60052),
            (*self._parse(f'{name}_EF_CONFIG_D'), 1),
            (*self._parse(f'{name}_EF_CONFIG_E'), 0)])

          chan.address, chan.dtype = self._parse(f'{name}_EF_READ_A')

        # Simplest case with no gain and offset
        elif chan.gain == 1 and chan.offset == 0 and not chan.make_zero:
          chan.address, chan.dtype = self._parse(name)

        # Setting the gain and offset if provided
        else:
          chan.write_at_open.extend([
            (*self._parse(f'{name}_EF_INDEX'), 1),
            (*self._parse(f'{name}_EF_CONFIG_D'), chan.gain),
            (*self._parse(f'{name}_EF_CONFIG_E'), chan.offset
             if not chan.make_zero else 0)])

          chan.address, chan.dtype = self._parse(f'{name}_EF_READ_A')

        self._channels_in.append(chan)

      # Digital to analog converters
      elif isinstance(name, str) and 'DAC' in name:
        chan = _Channel(name=name)
        chan.update(channel)

        chan.address, chan.dtype = self._parse(name)
        self._channels_out.append(chan)

      # Digital inputs and outputs
      elif isinstance(name, str) and 'IO' in name:
        chan = _Channel(name=name)
        chan.update(channel)

        chan.address, chan.dtype = self._parse(name)

        # Can be either input or output, the user has to specify
        if chan.direction:
          self._channels_out.append(chan)
        else:
          self._channels_in.append(chan)

      else:
        raise AttributeError(f"Invalid chan name: {name}")

    self.log(logging.DEBUG, f"Input channels: {self._channels_in}")
    self.log(logging.DEBUG, f"Output channels: {self._channels_out}")

    # Extracting the addresses and data types from all channels
    self._read_addresses = [chan.address for chan in self._channels_in]
    self._read_types = [chan.dtype for chan in self._channels_in]
    self._write_addresses = [chan.address for chan in self._channels_out]
    self._write_types = [chan.dtype for chan in self._channels_out]

    # These attributes will come in use later
    self._last_sent_val = [None for _ in self._write_addresses]

  def open(self) -> None:
    """Opens the Labjack, parses the commands to write at open, and sends
    them."""

    # Opening the Labjack
    self.log(logging.INFO, "Opening the connection to the Labjack")
    self._handle = ljm.openS(self._device, self._connection, self._identifier)

    # Gathering all the data to write at open
    self._write_at_open.extend(chain(*(chan.write_at_open for chan
                                       in chain(self._channels_in,
                                                self._channels_out))))

    # Parsing all the commands given as a tuple of two values
    self._write_at_open = [t if len(t) == 3 else (*self._parse(t[0]), t[1])
                           for t in self._write_at_open]

    # Getting the registers, data types and values in separate tuples
    if self._write_at_open:
      reg, types, values = tuple(zip(*self._write_at_open))
    else:
      reg = types = values = None

    # Finally, writing the commands to write at open
    if reg is not None:
      self.log(logging.DEBUG, f"Writing values {values} to addresses {reg}")
      ljm.eWriteAddresses(handle=self._handle,
                          numFrames=len(reg),
                          aAddresses=reg,
                          aDataTypes=types,
                          aValues=values)

  def make_zero(self, delay: float) -> None:
    """Overriding the method of the parent class, because the Labjack T7
    allows setting offsets directly on the board.

    Setting the offsets on the Labjack is slightly quicker than correcting the
    received values afterward.

    Args:
      delay: The delay during which the data should be acquired for determining
        the offset.
    
    .. versionadded:: 1.5.10
    """

    # No need to acquire data if no channel should be zeroed
    # Also, the IO channels shouldn't be zeroed
    if any(chan.make_zero and 'AIN' in chan.name
           for chan in self._channels_in):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:
        names, values = tuple(zip(*(
          (f"{chan.name}_EF_CONFIG_E", chan.offset + off)
          for off, chan in zip(self._compensations, self._channels_in)
          if chan.make_zero and 'AIN' in chan.name)))

        # Setting the offsets on the Labjack
        self.log(logging.DEBUG, f"Writing values {values} to registers "
                                f"{names}")
        ljm.eWriteNames(handle=self._handle,
                        numFrames=len(names),
                        aNames=names,
                        aValues=values)

        # Resetting the software offsets to avoid double compensation
        self._compensations = list()

  def get_data(self) -> List[float]:
    """Reads the signal on all pre-defined input channels, and returns the
    values along with a timestamp.."""

    return [time()] + ljm.eReadAddresses(handle=self._handle,
                                         numFrames=len(self._read_addresses),
                                         aAddresses=self._read_addresses,
                                         aDataTypes=self._read_types)

  def set_cmd(self, *cmd: float) -> None:
    """Sets the tension commands on the output channels.

    The given gain and offset are first applied, then the commands are clamped
    to the given limits. The commands are then written to the Labjack, only if
    they differ from the last written ones.
    """

    # First, applying the given gain and offsets to the commands
    cmd = [chan.gain * val + chan.offset
           for val, chan in zip(cmd, self._channels_out)]

    # Then, clamping the commands if limits were given
    cmd = [val if chan.limits is None
           else max(min(val, chan.limits[1]), chan.limits[0])
           for val, chan in zip(cmd, self._channels_out)]

    # Checking which values have to be updated and which are unchanged
    to_upd = [not val == prev for val, prev in zip(cmd, self._last_sent_val)]

    # Updating the list of last sent values
    self._last_sent_val = list(cmd)

    # No use continuing if there's nothing to update
    if any(to_upd):

      # Getting the addresses, data types and values to send
      addresses, types, values = tuple(zip(*(
        (addr, typ, val) for addr, typ, val, upd in
        zip(self._write_addresses, self._write_types, cmd, to_upd) if upd)))

      # Sending the commands
      if addresses:
        self.log(logging.DEBUG, f"Writing values {values} to addresses "
                                f"{addresses}")
        ljm.eWriteAddresses(handle=self._handle,
                            numFrames=len(addresses),
                            aAddresses=addresses,
                            aDataTypes=types,
                            aValues=values)

  def close(self) -> None:
    """Closes the connection to the Labjack."""

    if self._handle is not None:
      self.log(logging.INFO, "Closing the connection to the Labjack")
      ljm.close(self._handle)

  @staticmethod
  def _parse(name: str) -> Tuple[int, int]:
    """Wrapper around :meth:`ljm.nameToAddress` to make the code clearer."""

    return ljm.nameToAddress(name)



# coding: utf-8

import numpy as np

from .._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule('opencv-python')


class ApplyStrainToImage:
  """This class reshapes an image depending on input strain values. It is meant
  to simulate the stretching of a sample during a tensile test.

  Its main use case is for generating example scripts that do not require any
  hardware to run.
  
  .. versionadded:: 2.0.0
  """

  def __init__(self,
               image: np.ndarray) -> None:
    """Sets the base image and initializes the necessary objects to use."""

    self._img = image

    # Building the lookup arrays for the cv2.remap method
    height, width, *_ = image.shape
    orig_x, orig_y = np.meshgrid(range(width), range(height))
    # These arrays correspond to the original state of the image
    self._orig_x = orig_x.astype(np.float32)
    self._orig_y = orig_y.astype(np.float32)

    # These arrays are meant to be added to the original image ones
    # If added as is, they correspond to a 100% strain state in both directions
    self._x_strain = self._orig_x * width / (width - 1) - width / 2
    self._y_strain = self._orig_y * height / (height - 1) - height / 2

  def __call__(self, exx: float, eyy: float) -> np.ndarray:
    """Returns the reshaped image, based on the given strain values."""

    exx /= 100
    eyy /= 100

    # The final lookup table is the sum of the original state ones plus the
    # 100% strain one weighted by a ratio
    transform_x = self._orig_x - (exx / (1 + exx)) * self._x_strain
    transform_y = self._orig_y - (eyy / (1 + eyy)) * self._y_strain

    return cv2.remap(self._img, transform_x, transform_y, 1)



# coding: utf-8

from time import time
from typing import Optional, Tuple
import numpy as np
import logging
from  warnings import warn

from ..meta_camera import Camera
from ..._global import OptionalModule

try:
  from . import clModule as Cl
except (ImportError, ModuleNotFoundError):
  Cl = OptionalModule("clModule", "CameraLink module was not compiled. Please "
                                  "make sure /opt/SiliconSoftware/xxx/lib64 "
                                  "exists and reinstall Crappy")


class BaslerIronmanCameraLink(Camera):
  """This class can drive cameras over Camera Link through a Basler microEnable
  5 Ironman AD8 PoCL acquisition board.

  It is subclassed by the :class:`~crappy.lamcube.BiSpectral` and the
  :class:`~crappy.camera.cameralink.JaiGO5000CPMCL8Bits` Cameras. Not many
  settings can be accessed directly in Crappy, it is recommended to set them
  using a configuration file.

  Warning:
    This Camera relies on a custom-written C library that hasn't been tested in
    a long time. It might not be functional anymore. This Camera also requires
    proprietary drivers to be installed.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Cl_camera to BaslerIronmanCameraLink
  .. versionremoved:: 2.1.0
  """

  def __init__(self) -> None:
    """Adds the frame rate setting."""

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be "
         f"deprecated and removed from Crappy. Please contact the maintainers "
         f"if you still use this Camera.", FutureWarning)

    self._cap = None

    super().__init__()

    self.add_scale_setting("framespersec", 1, 200, self._get_framespersec,
                           self._set_framespersec)

  def open(self,
           num_device: int = 0,
           config_file: Optional[str] = None,
           camera_type: Optional[str] = None,
           **kwargs) -> None:
    """Reads the settings from the arguments or from the configuration file,
    sets them on the camera, and starts the acquisition.

    Args:
      num_device: The index of the camera to open if multiple cameras are
        connected, as an :obj:`int`.
      config_file: Path to the configuration file for the camera, as a
        :obj:`str`. Allows setting various parameters at once, and to store
        them in a persistent way.
      camera_type: The type of camera to drive, as a :obj:`str`.
      **kwargs: All the settings to set on the camera.
    
    .. versionadded:: 1.5.10
       explicitly listing the *num_device*, *config_file* and *camera_type*
       arguments
    .. versionchanged:: 2.0.0 renamed *numdevice* argument to *num_device*
    """

    # Reading the camera type from the config file, if applicable
    if camera_type is None and config_file is not None:
      self.log(logging.INFO, "Reading config file for getting the type of "
                             "camera")
      with open(config_file, 'r') as file:
        r = file.readlines()
      r = [s for s in r if s[:5] == "Typ='"]
      if r:
        camera_type = r[0][5:-3]

    if camera_type is None:
      raise AttributeError("No camera type or valid config file specified!")

    # Getting the data format from the kwargs
    if 'format' in kwargs:
      f = kwargs['format']
    # Getting the data format from the given camera type
    else:
      if camera_type[-1] == '8':
        f = Cl.FG_GRAY
      elif camera_type[-2:] == '16':
        f = Cl.FG_GRAY16
      elif camera_type[-2:] == '24':
        f = Cl.FG_COL24

      # Getting the data format from the configuration file
      else:
        if config_file:
          self.log(logging.WARNING, "Reading config file for getting the data "
                                    "format")
          with open(config_file, 'r') as file:
            r = file.readlines()
          r = [s for s in r if s[:10] == "FG_FORMAT="]
          if len(r) != 0:
            f = int(r[0].split('=')[1])
          else:
            raise ValueError("Could not determine the format")
        else:
          raise ValueError("Could not determine the format")

    # Opening the connection to the camera
    self.log(logging.INFO, "Initializing the communication with the camera")
    self._cap = Cl.VideoCapture()
    self._cap.open(num_device, camera_type, f)

    # Loading the parameters from the configuration file
    if config_file:
      self._cap.loadFile(config_file)

    self.set_all(**kwargs)

    # Starting the acquisition on the camera
    self.log(logging.INFO, "Starting acquisition")
    self._cap.startAcq()
    self._cap.set(Cl.FG_TRIGGERMODE, 1)
    self._cap.set(Cl.FG_EXSYNCON, 1)

  def get_image(self) -> Tuple[float, np.ndarray]:
    """Reads a frame from the camera and returns it as is."""

    ret, frame = self._cap.read()
    t = time()
    if not ret:
      raise IOError("Could not read camera")

    return t, frame

  def close(self) -> None:
    """Stops the acquisition, and releases the resources attributed to the
    camera."""

    if self._cap is not None:
      self.log(logging.INFO, "Stopping acquisition")
      self._cap.stopAcq()
      self.log(logging.INFO, "Closing the communication with the camera")
      self._cap.release()

  def _set_framespersec(self, val: float) -> None:
    self._cap.set(Cl.FG_FRAMESPERSEC, val)

  def _get_framespersec(self) -> float:
    return self._cap.get(Cl.FG_FRAMESPERSEC)

  def _set_h(self, val: int) -> None:
    self._cap.stopAcq()
    self._cap.set(Cl.FG_HEIGHT, val)
    self._cap.startAcq()

  def _set_w(self, val: int) -> None:
    self._cap.stopAcq()
    self._cap.set(Cl.FG_WIDTH, val)
    self._cap.startAcq()

  def _get_h(self) -> int:
    return self._cap.get(Cl.FG_HEIGHT)

  def _get_w(self) -> int:
    return self._cap.get(Cl.FG_WIDTH)
